
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.3. Last modification 13/03/08
*)
(**
interface.ml
@author Yann GLOUCHE
@version 1.1
@since 12/12/2005
*)
open Types;;
open Globals;;
open Utils;;
open Error_handler;;
open Utils_simulator;;
open Transcript;;
open Automate;;
open Variables;;
open Message_stack;;
open Unifytest;;
open Intruder;;
(*open Thread;;*)
open Tk;;

(** size of canva 
- in 800x480 resolution
let width = (ref (545.) : float Pervasives.ref);;
let height = (ref (500.) : float Pervasives.ref);;

- in 800x600 resolution 
let width = (ref (650.) : float Pervasives.ref);;
let height = (ref (690.) : float Pervasives.ref);;

- in 1280x1024 resolution 
let width = (ref (805.) : float Pervasives.ref);;
let height = (ref (670.) : float Pervasives.ref);;
 *)

let width = (ref (805.) : float Pervasives.ref);;
let height = (ref (670.) : float Pervasives.ref);;

let sim_version = "SPAN 1.3 -- Protocol Simulation"

let incoming_event_height=5;;
let past_event_height=5;;

(* For using files with pathes containing spaces we need an escape character *)

let os=Sys.os_type;;
let esc= if os="Unix" then "" else "\"";;



(** private coordinate system *)

let xmin = (ref (-5.) : float Pervasives.ref);;
let xmax = (ref (105.) : float Pervasives.ref);;
let ymin = (ref (-5.) : float Pervasives.ref);;
let ymax = (ref (105.) : float Pervasives.ref);;

let xscale = (ref (floor(!width /. float_of_int(abs(int_of_float(!xmax -. !xmin))))) : float Pervasives.ref);;
let yscale = (ref (floor(!height /. float_of_int(abs(int_of_float(!ymax -. !ymin))))) : float Pervasives.ref);;
    
(* ******************************************************* *)

(** graphicals objects *)

(** principal frame *)
let base = (Frame.create  root : Widget.frame Widget.widget);;

(** Menu bar *)
let bar = (Frame.create ~borderwidth:0 ~relief:`Flat base : Widget.frame Widget.widget);;
pack ~fill:`X [bar];;
(** frame for the canva *)
let canvasFrame = (Frame.create base : Widget.frame Widget.widget);;
(** canva for draw the MSC *)
let canvas = (Canvas.create ~height:(int_of_float !height) ~width:(int_of_float !width) ~background:selcol  ~relief:`Sunken
		  ~xscrollincrement:1 ~yscrollincrement:(int_of_float(round(1. *. !xscale)))  canvasFrame : Widget.canvas Widget.widget);;
let canvas_sb_v = (Scrollbar.create ~orient:`Vertical  canvasFrame : Widget.scrollbar Widget.widget);;
let canvas_sb_h = (Scrollbar.create ~orient:`Horizontal  canvasFrame : Widget.scrollbar Widget.widget);;
  scroll_link_canvas canvas_sb_v canvas "vertical";;
  scroll_link_canvas canvas_sb_h canvas "horizontal";;
  
(** frame for the buttons "Next step" "Previous step" "untyped_variables" and for for the incoming events and the past events*) 
let commandFrame = (Frame.create base : Widget.frame Widget.widget);;
(** frame for the incoming events and the past events *)
let infoFrame = (Frame.create ~width:20  commandFrame : Widget.frame Widget.widget);;

(** frame for the incoming events *)
let lstFrame = (Frame.create ~width:20  infoFrame : Widget.frame Widget.widget);;
(** label for the incoming events *)
let label_lst = (Label.create ~text:"Incoming events : " ~borderwidth:1 ~relief:`Ridge infoFrame : Widget.label Widget.widget);;
(** list box for the incoming events *)
let lst = (Listbox.create ~width:15 ~height:incoming_event_height ~background:selcol  ~selectmode:`Single lstFrame : Widget.listbox Widget.widget);;

(** scroll bar of lst *)
let lst_sb_v = (Scrollbar.create ~background: blue_light ~orient:`Vertical  lstFrame : Widget.scrollbar Widget.widget);;
let lst_sb_h = (Scrollbar.create ~background: blue_light ~orient:`Horizontal  lstFrame : Widget.scrollbar Widget.widget);;
  scroll_link lst_sb_v lst "vertical";;
  scroll_link lst_sb_h lst "horizontal";;

(** frame for the incoming events *)
let storyFrame = (Frame.create ~width:20  infoFrame : Widget.frame Widget.widget);;
(** label for the incoming events *)
let label_story = (Label.create ~text:"Past events : " ~borderwidth:1 ~relief:`Ridge infoFrame : Widget.label Widget.widget);;
(** list box for the incoming events *)
let story = (Listbox.create ~width:15 ~height:past_event_height ~background:defcol   ~selectmode:`Single storyFrame : Widget.listbox Widget.widget);;

(** scroll bar of story *)
let story_sb_v = (Scrollbar.create ~background: blue_light ~orient:`Vertical  storyFrame : Widget.scrollbar Widget.widget);;
let story_sb_h = (Scrollbar.create ~background: blue_light ~orient:`Horizontal  storyFrame : Widget.scrollbar Widget.widget);;
  scroll_link story_sb_v story "vertical";;
  scroll_link story_sb_h story "horizontal";;
  
(** frame for the buttons "Next step" "Previous step" "untyped_variables" *)
let buttonFrame = (Frame.create  ~relief:`Ridge  ~borderwidth:1 commandFrame : Widget.frame Widget.widget);;

pack ~fill:`Both ~expand:true [base];;

pack ~fill:`Both ~side:`Top[buttonFrame];;
pack ~side:`Left ~fill:`Y [lst_sb_v];
pack ~side:`Bottom ~fill:`X [lst_sb_h];
pack ~side:`Top ~fill:`Both ~expand:false [label_lst];;
pack ~side:`Top ~fill:`Both ~expand:true [lst];;
pack ~side:`Top ~fill:`Both ~expand:true [lstFrame];;
pack ~side:`Left ~fill:`Y [story_sb_v];
pack ~side:`Bottom ~fill:`X [story_sb_h];
pack ~side:`Top ~fill:`Both ~expand:false [label_story];;
pack ~side:`Top ~fill:`Both ~expand:true [story];;
pack ~side:`Top ~fill:`Both ~expand:true [storyFrame];;

pack ~side:`Top ~fill:`Both ~expand:true [infoFrame];;
pack ~side:`Left ~fill:`Both ~expand:true [commandFrame];;

pack ~side:`Right ~fill:`Y [canvas_sb_v];
pack ~side:`Bottom ~fill:`X [canvas_sb_h];
pack ~side:`Top ~fill:`Both ~expand:true [canvas];;
pack ~side:`Left ~fill:`Both ~expand:true [canvasFrame];;

(** remove the graphical objects (arrow, text, message) created by the execution of the previous step of the protocol
@return unit
*)	
let remove_graphics_obj (() : unit) : unit = 
	if((!tag_number) >=0) then
	(Canvas.delete canvas [`Tag ("tag"^(string_of_int (!tag_number)))]);;

(** remove the graphical objects associated with the parameter tag
@param tag string
@return unit
*)
let remove_graphics_obj_with_tag_value (tag : string) : unit = 
	(Canvas.delete canvas [`Tag (tag)]);;

(** remove the graphical objects associated with the two parameters tag_num and tag_string
@param tag_num int
@param tag_string string
@return unit
*)	
let remove_graphics_obj_with_tag (tag_num : int) (tag_string : string) : unit = 
	(Canvas.delete canvas [`Tag (tag_string^(string_of_int tag_num))]);;

(** the distance between two arrows*)
let yDelta = (ref 10. : float Pervasives.ref);;
(** converse a abscissa value in the private coordinate system to the real coordinate system of the canva 
@param x float
@return int
*)
let xconv (x : float) : int = int_of_float ((x -. !xmin)*.(!xscale));;
(** converse a Y-axis value in the private coordinate system to the real coordinate system of the canva 
@param y float
@return int
*)
let yconv (y : float) : int = int_of_float ((!ymax -. y)*.(!yscale));;
(** draw between (x1, y1) and (x2, y2) with the color col and the tag tag_number
@param x1 float
@param y1 float
@param x2 float
@param y2 float
@param col Tk.color
@param tag_number int
@return unit
*)
let draw_line (x1 : float) (y1 : float) (x2 : float) (y2 : float) (col : Tk.color) (tag_number : int) : unit = 
				    let xc1 = xconv x1 in
				    let yc1 = yconv y1 in
				    let xc2 = xconv x2 in
				    let yc2 = yconv y2 in
				    ignore(Canvas.create_line ~xys:[xc1, yc1 ; xc2, yc2]  ~width:4 ~tags:[("tag"^(string_of_int tag_number))] ~fill:col canvas);;

(** 
For the variable monitoring

draw a rectangle around the point (x,y) the tags values tag_number1 and tag2.
@param x float
@param y float
@param tag_number1 int
@param tag2 string
@return unit
*)				    
let draw_click_zone (x : float) (y : float) (tag_number1 : int) (tag2 : string) : unit = 
				    let xc1 = xconv (x -. 0.25) in
				    let yc1 = yconv (y +. 0.55) in
				    let xc2 = xconv (x +. 0.25) in
				    let yc2 = yconv (y -. 0.55) in
				    ignore(Canvas.create_rectangle ~x1:xc1 ~y1:yc1  ~x2:xc2 ~y2:yc2 ~width:1 ~tags:[("tag"^(string_of_int tag_number1));tag2] ~fill:clicol ~outline:violet canvas);;
		
(** if !message_emission=true then the animator displays all the messages send in the network for each transitions *)		    
let message_emission = (ref false : bool Pervasives.ref);;

(** if !message_sender=true then the animator displays the pattern of the message send in the network for each transitions *)		    
let message_sender = (ref false : bool Pervasives.ref);;
(** if !message_receiver=true then the animator displays the pattern of the message receiver in the network for each transitions *)		    
let message_receiver = (ref false : bool Pervasives.ref);;
(** if !message_real=true then the animator displays the message send by an agent and receive by another*)		    
let message_real = (ref true : bool Pervasives.ref);;
(** if !message_principal_sender=true then the animator displays the pattern of the message send in the network for each transitions in the middle of the arrow *)		    
let message_principal_sender = (ref false : bool Pervasives.ref);;

(** draw a text at the position (x,y) with color col with tag value tag
@param x float
@param y float
@param text string
@param col Tk.color
@param tag string list
@return unit
*)
let draw_text (x : float) (y : float) (text : string) (col : Tk.color) (tag : string list) (a : Tk.anchor) : unit = 
				 let xc = xconv x in
				 let yc = yconv y in 
				 ignore(Canvas.create_text ~x:xc ~y:yc~font:"textfont" ~fill:col ~width:(xconv 100.) ~anchor:a ~tags:tag ~text:text  canvas);;
				 
(** 
For the variable monitoring

draw a text at the position (x,y) with color col with tag value tag_number and tag2
@param x float
@param y float
@param text string
@param col Tk.color
@param tag_number int
@param tag2 string
@return unit
*)	 
let draw_text_left (x : float) (y : float) (text : string) (col : Tk.color) (tag_number : int) (tag2 : string) : unit = 
				 let xc = 5+(xconv x) in
				 let yc = yconv y in 
				 ignore(Canvas.create_text ~x:(xc+3) ~y:yc ~font:"guifont" ~fill:col ~width:(xconv 100.) ~tags:[("tag"^(string_of_int tag_number));(tag2^(string_of_int tag_number))] ~anchor:`W ~text:text  canvas);
				 ignore(Canvas.create_line ~xys:[xc,(yc-(int_of_float (!yscale *. 0.5 *. !yDelta -. 0.1))) ; xc,(yc+(int_of_float(!yscale *. 0.5 *. !yDelta -. 0.1)))] ~width:2 ~tags:[("tag"^(string_of_int tag_number));(tag2^(string_of_int tag_number))] ~fill:col canvas);
;;

(** 
For display the number of the step of protocol

draw a text at the position (x,y) with color col with tag value tag_number and tag2
@param x float
@param y float
@param text string
@param col Tk.color
@param tag_number int
@param tag2 string
@return unit
*)
let draw_step (x : float) (y : float) (text : string) (col : Tk.color) (tag_number : int) (tag2 : string) : unit = 
				 let xc = xconv x in
				 let yc = yconv y in 
				 ignore(Canvas.create_text ~x:xc ~y:yc ~font:"guifont" ~fill:col ~width:(xconv 100.) ~tags:[("tag"^(string_of_int tag_number));(tag2^(string_of_int tag_number))] ~anchor:`E ~text:text  canvas);
;;

(** role_var_monitoring contains all the name the variables of the roles for which it is possible to display the value
- the key is the role
- the value is the name of the variables of the role
*)			 
let role_var_monitoring = (Hashtbl.create 600 : (string, string) Hashtbl.t);;

(** var_monitoring contains all the name the variables of the roles for which the user wants to display the value
- the key is the role
- the value is the name of the variables of the role
*)
let var_in_monitor = (Hashtbl.create 600 : (string, string) Hashtbl.t);;

(** the infos_zones is the zones used for the variable monitoring is a click of all the click_zones associated with the Y-axis position of the information
*)
let infos_zones = (ref []: ((float * float) * float list) list Pervasives.ref);;

(** compute the number of the values of var displayed on the MSC associated with the click zone (abs,ord).
@param abs float
@param ord float
@return int
*)
let find_number_of_var_in_monitor (abs : float) (ord : float) :int = 					
						if(List.mem true (List.map (fun ((x,y),_) -> x=abs && y=ord) (!infos_zones) ))
						then (
							let z = List.assoc (abs,ord) (!infos_zones) in
							int_of_float ( 0. +. ((max_l z) -. (min_l z)) /. !yDelta)
						)
						else 0;;
						
(** Test if the Y-axis position ord is contained in a infos_zones (a zone for the display the values of variables)
@param ord float
@return bool
*)
let in_zone (ord : float) : bool =
	let rec aux l =
		match l with [] -> false
			|((_,z)::ls) -> (
					let mini = min_l z in
					let maxi = max_l z in
					if ( (mini <= ord && ord <= maxi) )
					then true
					else aux ls 
				)
	in aux (!infos_zones)
;;
(**
test if point (x,y) is contained in a zone of click
@param x is the abscissa value of the point (x,y)
@param y is the Y-axis value of the point (x,y)
@return boolean, true if (x,y) is in a zone of click, else false
*)
let in_click_zone (x : float) (y : float) : bool =
	List.mem true (List.map (fun ((click_x,click_y),_) -> 
					(x = click_x && click_y = y) ) 
	(!infos_zones))	
;;
(**
test if a zone of click with the Y-axis value equals to y, exists
@param y is the Y-axis value of the zone of click that we search with this function
@return boolean, true if the zone of exists, else false
*)
let in_click_zone_y (y : float) : bool =
	List.mem true (List.map (fun ((_,click_y),_) -> 
					click_y = y ) 
	(!infos_zones))	
;;
(**
 compute the number of common line between the zone associates with the click zone (x,y), and all the others zones (the zones for the display the values of variables : "variable monitoring") with the Y-axis value equals to "y".
 @param x is an abscissa value
 @param y is an Y-axis value
 @return a number of lines
*)
let number_of_common_line_between_all_zones (x : float) (y : float) : int =
	let rec aux infos_zones res =
		match infos_zones with [] -> (*prerr_string "|= ";prerr_float((res /. !yDelta)); prerr_string "=| ";*)(int_of_float (res /. !yDelta)) 
						|(((click_x,click_y),z)::ls) ->
							if(click_x<>x && click_y=y)
							then aux ls (max res ((max_l z) -. (min_l z))) 
							else aux ls res; 
	in aux (!infos_zones) 0.
;;
(**
remove the infos zone for which the click zone position is (x,y)
@param x in an abscissa value
@param y is an Y-axis value
@return unit
*)
let del_zone (x : float) (y : float) : unit =
	let zone_length = ref 0. in
	let rec aux0 l res =
		match l with [] -> List.rev res
			|(((click_x,click_y),zone)::ls) -> 	
						if(click_y=y && click_x=x)
						then  (zone_length := ((max_l zone) -. (min_l zone)) ; aux0 ls res)
						else aux0 ls (((click_x,click_y),zone)::res);
	in infos_zones:=(aux0 (!infos_zones) []);

	
	let dist = !zone_length -. (min !zone_length (((float_of_int (number_of_common_line_between_all_zones x y)) -. 0.)  *. !yDelta)) in
	let rec aux l_infos_zones res =
		match l_infos_zones with [] -> List.rev res
			|(((absX,ordY),points_list)::ls) ->
				if(ordY <= y)
				then aux ls (((absX,(ordY +. dist)),(List.map (fun p -> p +. dist) points_list))::res)
				else aux ls (((absX,ordY),points_list)::res)
	in 
	
	let lock = ref true in
	List.iter
	(fun ((role_x,tag_pos),zone) ->
	if (role_x <> x && tag_pos=y && ((max_l zone) -. (min_l zone)) >= (!zone_length))
	then lock:=false)
	(!infos_zones);	
	if(!lock) then(
			infos_zones:= aux (!infos_zones) []
	)	
;;
(**
remove all the infos zones for which the Y-axis value of the click zone is equals to "y"
@param y is an Y-axis value
@return unit
*)
let del_zone_y (y : float) : unit =
	let dist = ref 0. in
	let rec aux l res =
		match l with [] -> List.rev res
			|(((click_x,click_y),zone)::ls) -> 	
						if(click_y=y)
						then (	dist:= max !dist ((max_l zone) -. (min_l zone)) ; 
							aux ls res)
						else (aux ls (((click_x,click_y),zone)::res));
	in infos_zones:=(aux (!infos_zones) []);

	
	let rec aux0 l_infos_zones res =
		match l_infos_zones with [] -> List.rev res
			|(((absX,ordY),points_list)::ls) ->
				if(ordY <= y)
				then aux0 ls (((absX,(ordY +. !dist)),(List.map (fun p -> p +. !dist) points_list))::res)
				else aux0 ls (((absX,ordY),points_list)::res)
	in infos_zones:= aux0 (!infos_zones) [];

;;
(**
add a zone for which the click zone position is equal to (x,y), and the list of Y-axis position of the info zones is y.
@param x in an abscissa value
@param y is an Y-axis value
@return unit
*)
let add_in_zone (abs : float) (ord : float) (y : float list) : unit = 
	let min_l l = List.fold_left min (car l) (cdr l) in
	let max_l l = List.fold_left max (car l) (cdr l) in
	let zone_length = (max_l y) -. (min_l y) in
	let dist = zone_length -. (min zone_length ( ( (float_of_int (number_of_common_line_between_all_zones abs ord)) -. 0.) *. !yDelta )) in
	let rec aux l_infos_zones res =
		match l_infos_zones with [] -> List.rev res
			|(((absX,ordY),points_list)::ls) ->
				if(ordY <= ord)
				then aux ls (((absX,(ordY -. dist)),(List.map (fun p -> p -. dist) points_list))::res)
				else aux ls (((absX,ordY),points_list)::res)
	in 	
	let lock = ref true in
	List.iter
	(fun ((role_x,tag_pos),zone) ->
	if (role_x <> abs && tag_pos=ord && ((max_l zone) -. (min_l zone)) >= zone_length)
	then lock:=false)
	(!infos_zones);	
	if(!lock) then ( 
			infos_zones:= aux (!infos_zones) []
		);

	let ord_zone = if(!lock) then (ord -. dist) else ord in
	if not(List.mem (abs,ord_zone) (List.map (fun ((x,z),_) -> (x,z) ) (!infos_zones)))
	then infos_zones := ((abs,ord_zone),y)::(!infos_zones)
	else (
		let y2 = List.assoc (abs,ord_zone) (!infos_zones) in
		let y3 = y2@y in
		infos_zones := ((abs,ord_zone),[(min_l y3) ; (max_l y3)])::(remove ((abs,ord_zone),y2) (!infos_zones))
		
	)
;;	

(** The number of the agents in the protocol execution *)
let nbAgents = (ref 0 : int Pervasives.ref);;
(** A marge at the top of the MSC*)
let yMarge = (ref 5. : float Pervasives.ref);;
(** The Y-axis position of the transition during the protocol execution*)
let yOrd = (ref 0.0 : float Pervasives.ref);;
(** Initializationof yOrd*)
let yOrd_init (() : unit) : unit = yOrd:=((!ymax)-. (!yMarge) -. (!yDelta));;
yOrd_init ();;
(** Adjust the value of yOrd after the execution of a transition*)
let yAction (() : unit) : float = 
	let res = !(yOrd) in
	let rec aux () =
		yOrd := (!yOrd)-.(!yDelta);
		if(in_zone (!yOrd)) then aux ()
	in aux ();
	Canvas.configure canvas ~scrollregion:((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) ),0,(xconv ((float_of_int(!nbAgents + 1)) *. 10. +. 5. )) + (5*(get_max_length_message())), (yconv !yOrd));
	res 
;;

(** Adjust the value of yOrd when the user come back to the previous step of the execution of the protocol *)
let previous_yAction (() : unit) : unit =
	let rec aux lock =
		if((!yOrd)< ((!ymax)-. !yMarge -. !yDelta))
		then(
			yOrd := (!yOrd)+.(!yDelta);
			if(lock && in_click_zone_y (!yOrd)) then (let old_y = (!yOrd) in aux false ; del_zone_y old_y)
			else if(in_zone (!yOrd) ) then ( aux false )
			else if not(lock) then yOrd := (!yOrd)-.(!yDelta);
		)
	in aux true;
	Canvas.configure canvas ~scrollregion:((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) ),0,(xconv ((float_of_int(!nbAgents + 1)) *. 10. +. 5. )) + (5*(get_max_length_message())), (yconv !yOrd));
;;


(** the abscissa of the agents *)
let role_position = (Hashtbl.create 600 : (string, float) Hashtbl.t);;

(**
draw an agent of the protocol
@param x_pos float
@param text string
@param col Tk.color
@return unit
*) 
let draw_plot (x_pos : float) (text : string) (col : Tk.color) : unit = 
   draw_line x_pos ((!ymax)*. (-1000.)) x_pos (!ymax -. 5.) col (-1) ;
   draw_line (x_pos -. 3.5) ((!ymax)-. 5.) (x_pos +.3.5) ((!ymax)-. 5.) col (-1) ;
   draw_text x_pos ((!ymax)-. 3.) text col [("tag"^(string_of_int (-1)))] `Center;
   Hashtbl.add role_position text x_pos;;

(** draw an arrow beteween  (x1, y) and (x2,y) with the color col_line, for symbolise an execution of a step of the protocol, with the texts message_sender, real_message and message_receiver with the color col_message, and with the text message_emission with the color col_emission. message_emission is all the messages send the network during the execution of the step, and txt is message send by an agent to anoter for the execution of the transition
@param x1 float
@param x2 float
@param y float
@param col_line Tk.color
@param message_receiver string
@param real_message string
@param message_sender string
@param col_message Tk.color
@param message_emission string
@param col_emission Tk.color
@return unit
*)
let draw_row (x1 : float) (x2 : float) (y : float) (col_line : Tk.color) (message_receiver : string)(real_message : string)(message_sender : string) (col_message : Tk.color) (message_emission : string) (col_emission : Tk.color) : unit = 
     draw_line x1 y x2 y col_line (!tag_number) ;
     if (x1 < x2)
     then (draw_line x2 y (x2 -. 1.) (y +. 1.) col_line (!tag_number) ; 
           draw_line x2 y (x2 -. 1.) (y -. 1.) col_line (!tag_number) ;
	   draw_text (x1 -. 0.3) (y +. 0.2) message_sender col_message [("tag"^(string_of_int (!tag_number)));"message_sender"] `E;
	   draw_text (x2 +. 0.3) (y +. 0.2) message_receiver col_message [("tag"^(string_of_int (!tag_number)));"message_receiver"] `W;
	   )
     else (draw_line x2 y (x2 +. 1.) (y +. 1.) col_line (!tag_number) ; 
           draw_line x2 y (x2 +. 1.) (y -. 1.) col_line (!tag_number) ;
	   draw_text (x1 +. 0.3) (y +. 0.2) message_sender col_message [("tag"^(string_of_int (!tag_number)));"message_sender"] `W;
	   draw_text (x2 -. 0.3) (y +. 0.2) message_receiver col_message [("tag"^(string_of_int (!tag_number)));"message_receiver"] `E;
	   );
     draw_text ((min x1 x2)+.((sqrt((x1 -. x2) *. (x1 -. x2)))/. 2.)) (y+. 2.7) real_message col_message [("tag"^(string_of_int (!tag_number)));"message_real"] `Center;
     draw_text ((min x1 x2)+.((sqrt((x1 -. x2) *. (x1 -. x2)))/. 2.)) (y+. 2.7) message_sender col_message [("tag"^(string_of_int (!tag_number)));"message_principal_sender"] `Center;
     draw_text x1 (y-. 2.) message_emission col_emission [("tag"^(string_of_int (!tag_number)));"emission"] `Center;
     ;;
     
(** draw an arrow between  then agentA and the agentB with the color col_line, for symbolise an execution of a step of the protocol, with the text sender_pattern, real_message, and message_receiver with the color col_txt, and with the text message2 with the color col_txt2. message2 is all the messages send the network during the execution of the step, and sender_pattern is message send by an agent to anoter for the execution of the transition.
receive_pattern is the pattern of the message receive by an agent, and message is the pattern of the message send by an agennt to another. real_message is the instanciation of the pattern of message_sender
@param agentA string
@param agentB string
@param col Tk.color
@param sender_pattern string
@param real_message string
@param receive_pattern string
@param col_txt Tk.color
@param message2 string
@param col_txt2 Tk.color
@return unit
*)
let draw_action (agentA : string) (agentB : string) (col : Tk.color)(sender_pattern : string)(real_message : string)(receive_pattern : string)(col_txt : Tk.color) (message2 : string) (col_txt2 : Tk.color) =
	let x_posA = Hashtbl.find role_position agentA in
	let x_posB = Hashtbl.find role_position agentB in
	let old_yOrd = !yOrd in
	draw_row x_posA x_posB (yAction()) col receive_pattern real_message sender_pattern col_txt message2 col_txt2;
	draw_step ((!xmax)-. 0.1) (old_yOrd +. 5.) ("Step"^(string_of_int !tag_number)^".") col (!tag_number) "step";
	Hashtbl.iter
	(fun role_name role_x ->
		if( (Hashtbl.find_all var_in_monitor role_name)<>[] ) 
		then draw_click_zone role_x old_yOrd (!tag_number) (role_name^"click_zone")
	)
	role_position;
	if(!message_emission) 
	then (Canvas.configure_text canvas (`Tag "emission") ~state:(`Disabled) )
	else (Canvas.configure_text canvas (`Tag "emission") ~state:(`Hidden));
	if(!message_sender) 
	then (Canvas.configure_text canvas (`Tag "message_sender") ~state:(`Disabled) )
	else (Canvas.configure_text canvas (`Tag "message_sender") ~state:(`Hidden));
	if(!message_receiver) 
	then (Canvas.configure_text canvas (`Tag "message_receiver") ~state:(`Disabled) )
	else (Canvas.configure_text canvas (`Tag "message_receiver") ~state:(`Hidden));
	if(!message_real) 
	then (Canvas.configure_text canvas (`Tag "message_real") ~state:(`Disabled))
	else (Canvas.configure_text canvas (`Tag "message_real") ~state:(`Hidden));
	if(!message_principal_sender) 
	then (Canvas.configure_text canvas (`Tag "message_principal_sender") ~state:(`Disabled))
	else (Canvas.configure_text canvas (`Tag "message_principal_sender") ~state:(`Hidden));
;;

(**
initialize the X-axis
@param nb_agents int
@return unit
*)
let xinit_coord (nb_agents : int) : unit = 
 xmax := (min (float_of_int(nb_agents-1)) 9.) *. 10. +. 5.;
 xscale := (!width /. float_of_int(abs(int_of_float(!xmax -. !xmin))));
 nbAgents:=nb_agents;
 Canvas.configure canvas ~scrollregion:((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) ),0,(xconv ((float_of_int(!nbAgents + 1)) *. 10. +. 5. )) + (5*(get_max_length_message())), (yconv !yOrd))
;;

(* %%%%%%%%%%%%%%%%%%%%%%%%%% *)



(*
   This module defines an interface with the system output.
   The functions work on data structures returned by the instantation module.
   They perform last formatting steps of the IF sections and print them on stdin.
 *)

(*
   When a rule is printed, we may have to add some [Contains] fact. We build the list
   of these facts as the states are printed.
  \begin{center}
  type:
   [val contains_term_list : (Types.if_value_type * Types.if_value_type) list ref]
  \end{center}
*)
let contains_term_list = ref ([]: (if_value_type * if_value_type) list);;

(* 
   The signature section contains state signatures such as:
    \begin{center}
   [state_Alice(agent,agent,public_key,public_key,nat,{3,1,2},nonce,nonce,nat) -> fact]
   \end{center}

   In order to print a signature only once, we store it in the following list.
  \begin{center}
  type:
   [val roles_printed : string list ref]
  \end{center}
*)
let roles_printed = ref ([] : string list);;


(*
  \begin{center}
  type:
   [val list_to_string : ('a -> string) -> string -> 'a list -> string]\\
  type:
   [val print_list : ('a -> string) -> string -> 'a list -> unit]\\
  type:
   [val prerr_list : ('a -> string) -> string -> 'a list -> unit]
  \end{center}
*)

let list_to_string (f : 'a->string) (separator : string) : 'a list -> string =
  let rec list_to_string_rec = function
      [] -> ""
    | [h] -> f h
    | h::t -> (f h)^separator^(list_to_string_rec t)
  in
  list_to_string_rec
;;

let print_list (f : 'a -> string) (separator : string) (l : 'a list) : unit =
  prout_string (list_to_string f separator l)
;;

let prerr_list (f : 'a -> string) (separator : string) (l : 'a list) : unit =
  prerr_string (list_to_string f separator l)
;;

(*
  \begin{center}
  type:
   [val list2_to_string : ('a -> string) -> string -> 'a list -> string]
  \end{center}
*)
let list2_to_string (f : 'a -> string) (separator : string) : 'a list -> string =
  let rec list2_to_string_rec = function
      [] -> ""
    | [h] -> f h	
    | h::t ->
 	"pair("^(f h)^separator^(list2_to_string_rec t)^")"
  in
  list2_to_string_rec
;;


(*
   [print_term_type] is used for section types or signature to print
   recursively the types of terms.
  \begin{center}
  type:
   [val term_type_to_string : Types.basic_type Types.term_structure -> string]\\
  type:
   [val prerr_term_type : Types.basic_type Types.term_structure -> unit]
  \end{center}
*)
let rec term_type_to_string : Types.basic_type Types.term_structure -> string = function
    Base(term) | Prime(term) ->
      (match term with
	Agent -> "agent"
      | Public_key -> "public_key"
      | Symmetric_key -> "symmetric_key"
      | Text -> "nonce"
      | Protocol_id -> "protocol_id"
      | Message -> "message"
      | Channel -> "channel"
      | Channel_dy -> "channel(dy)"
      | Channel_ota -> "channel(ota)"
      | Bool -> "bool"
      | Hash_func -> "hash_func"
      | Nat -> "nat"
(*      | Pointer -> "pointer"*)
      | Enum(list) ->
	  "{"
	  ^(List.fold_left
	      (fun s num ->
	      	if not (num = List.hd list) then
	      	  s^","^(string_id#get_name num)
	      	else
		  s^(string_id#get_name num))
	      "" list)
	  ^"}")
	
  | Function(term,list)->
      (match list with
	[] -> "unit"
      | _ ->
	  (List.fold_left
 	    (fun s t -> 
	      s^(term_type_to_string t)^" -> ")
	    "" list)
	  ^(term_type_to_string term))
      
(*  | Lst(list)->
      (List.fold_left
	 (fun s t ->
	   s^(term_type_to_string t))
	 "" list)
      ^" list"*)
      
  | Set([])->
     "set"
  | Set(list)->
     (term_type_to_string (List.hd list))^" set"
      
  | Pair(term1,term2)->
      "pair("
      ^(term_type_to_string term1)^","^(term_type_to_string term2)
      ^")"
        
  | _ -> "nonce"
;;

let prerr_term_type (t : Types.basic_type Types.term_structure) : unit =
  prerr_string (term_type_to_string t)
;;


(*
  \section{Generic printing functions}

  \subsection{Printing [if_value_type] terms}
  \begin{center}
  type:
   [val if_value_to_string :
   (Types.if_value_type -> string) -> Types.if_value_type -> string]\\
  type:
   [val print_if_value :
   (Types.if_value_type -> string) -> Types.if_value_type -> unit]\\
  type:
   [val prerr_if_value :
   (Types.if_value_type -> string) -> Types.if_value_type -> unit]
  \end{center}
*)
let rec if_value_to_string (atom_to_string : (Types.if_value_type -> string)) (if_value : Types.if_value_type) : string =
  let if_value_to_string_aux = if_value_to_string atom_to_string in
  match if_value with
    If_address _ | If_const _ ->
      atom_to_string if_value
  | If_pair(term1,term2) ->
      "pair("^(if_value_to_string_aux term1)^","
      ^(if_value_to_string_aux term2)^")"
  | If_scrypt(term1,term2) ->
      "scrypt("^(if_value_to_string_aux term2)^","
      ^(if_value_to_string_aux term1)^")"
  | If_crypt(term1,term2) ->
      "crypt("^(if_value_to_string_aux term2)^","
      ^(if_value_to_string_aux term1)^")"
  | If_inv(term) ->
      "inv("^(if_value_to_string_aux term)^")"
  | If_function(term1,termlist) ->
      begin
	match term1 with
	    If_const(fun_id) ->
	      let fun_string = (string_id#get_name fun_id) in
		if (fun_string = "exp") || (fun_string = "xor") then
		  fun_string^"("
		  ^(list_to_string if_value_to_string_aux "," termlist)
		  ^")"
		else
		  "apply("
		  ^(if_value_to_string_aux term1)
		  ^","
		  ^(list2_to_string if_value_to_string_aux "," termlist)
		  ^")"
	  | _ -> 
	      "apply("^(if_value_to_string_aux term1)^","
	      ^(list2_to_string if_value_to_string_aux "," termlist)^")"
      end
  | If_cons(term1,term2) ->
      "cons("^(if_value_to_string_aux term1)^","
      ^(if_value_to_string_aux term2)^")"
  | If_delete(term1,term2) ->
      "delete("^(if_value_to_string_aux term1)^","
      ^(if_value_to_string_aux term2)^")"
  | If_set(termlist) ->
      "set("^(list_to_string if_value_to_string_aux "," termlist)^")"
(*  | If_lst(termlist) ->
      "list("^(list_to_string if_value_to_string_aux "," termlist)^")"*)
;;

let print_if_value (atom_fn : (Types.if_value_type -> string)) (if_value : Types.if_value_type) : unit =
  prout_string (if_value_to_string atom_fn if_value)
;;

let prerr_if_value (atom_fn : (Types.if_value_type -> string)) (if_value : Types.if_value_type) : unit =
  prerr_string (if_value_to_string atom_fn if_value)
;;

(*
  \begin{center}
  type:
   [val atom_ground_value_to_string : Types.if_value_type -> string]
  \end{center}
*)
let rec atom_ground_value_to_string : Types.if_value_type -> string = function
  | If_address gid-> 
      (*
	 In the case of a variable, we get its initial value from
	 the [Globals.value_table] variable. 
       *)
      let map_id = memory_map#get_map gid in
      let if_value = value_table#get_value map_id in
      if if_value = If_address gid then
	(let vid=global_var_id#get_name_of gid in
	 let name = string_id#get_name vid in
	 Error_handler.displayWarning 4 ("Interface.atom_ground_value_to_string."^name);
	 "")
      else
	if_value_to_string atom_ground_value_to_string if_value (* \fix{is it possible to reuse [atom_value_to_string] } *)

  | If_const strid ->
      begin
	(string_id#get_name strid)^
	(*
	  If the constant is a pointer to a set, we also have to print the contains terms according to
	  the value of this set. To this end, we add the right terms in the [contains_term_list] variable.
	*)
	(try
	  match (Globals.type_table#get_type strid) with
(*	    Base(Pointer)
	  | Lst(_)*)
	    Set(_) ->
	      begin
		let pointer_global_var_id = Globals.global_var_id#get_id_of 0 strid in
		let pointer_mmap_id = 
		  Globals.memory_map#get_map pointer_global_var_id in
		let if_value_map = Globals.value_table#get_value pointer_mmap_id in 
		match if_value_map with
(*		  If_lst l *)
		  If_set l ->
		    contains_term_list := 
		      (List.map 
			 (fun element -> 
(*			   if (!debug_level >= 3) then
			     begin
			       prerr_string ("element in the set "
					     ^(Globals.string_id#get_name strid)
					     ^": ");
			       prerr_if_value atom_ground_value_to_string element;
			       prerr_newline();
			     end;*)
			   (element,If_const(strid))) 
			 l)
		      @ !contains_term_list;
		    ""
		| _ -> 
		    prerr_endline ("Interface.atom_ground_value_to_string: found a pointer to something else than a set or a list");
		    raise Not_found
	      end;
	  | _ -> ""
	with
	  _  -> "")
      end
     (*
        This last case should not be encountered as long as the [print_atom_ground_value]
	is called only in the previous cases.
      *)
  | _ -> prerr_endline "Interface.atom_ground_value_to_string: considering a non-atomic term as atomic"; ""
;;


(*
  \begin{center}
  type:
   [val ground_if_value_to_string : Types.if_value_type -> string]\\
  type:
   [val print_ground_if_value : Types.if_value_type -> unit]\\
  type:
   [val prerr_ground_if_value : Types.if_value_type -> unit]
  \end{center}
*)
let ground_if_value_to_string : Types.if_value_type -> string =
  if_value_to_string atom_ground_value_to_string
;;
let print_ground_if_value : Types.if_value_type -> unit =
  print_if_value atom_ground_value_to_string
;;
let prerr_ground_if_value : Types.if_value_type -> unit =
  prerr_if_value atom_ground_value_to_string
;;


(*
  \subsection{Printing [term] terms}
*)


(*
   This function prints input terms in IF format.
  \begin{center}
  type:
   [val term_to_string : Types.atoms_type Types.term_structure -> string]\\
  type:
   [val print_term : Types.atoms_type Types.term_structure -> unit]\\
  type:
   [val prerr_term : Types.atoms_type Types.term_structure -> unit]
  \end{center}
*)
let rec term_to_string : Types.atoms_type Types.term_structure -> string = function
    Pair(term1,term2) ->
      "pair("^(term_to_string term1)
      ^","^(term_to_string term2)^")"

  | Scrypt(term1,term2) ->
      "scrypt("^(term_to_string term2)
      ^","^(term_to_string term1)^")"
      
  | Crypt(term1,term2) ->
      "crypt("^(term_to_string term2)
      ^ ","^(term_to_string term1)^")"

  | Inv(term) ->
      "inv("^(term_to_string term)^")"

  | Function(term1,termlist) ->
      begin
	match term1 with
	    Base(Const(fun_id)) ->
	      let fun_string = (string_id#get_name fun_id) in
		if (fun_string = "exp") || (fun_string = "xor") then
		  fun_string^"("
		  ^(list_to_string term_to_string "," termlist)
		  ^")"
		else
		  "apply("^(term_to_string term1)
		  ^","^(list2_to_string term_to_string "," termlist)
		  ^")"
	  | _ -> 
	      "apply("^(term_to_string term1)
	      ^","^(list2_to_string term_to_string "," termlist)
	      ^")"
      end

  | Cons(term1,term2) ->
      "contains("^(term_to_string term1)
      ^","^(term_to_string term2)^")"

  | Delete(term1,term2) -> (* to be contains... *)
      "delete("^(term_to_string term1)
      ^","^(term_to_string term2)^")"

  | Set(termlist) ->
      "set("^(list_to_string term_to_string "," termlist)^")"
(*BUG??? 15/03/2005
      list2_to_string term_to_string "," termlist *)

(*  | Lst(termlist) ->
      "list("^(list_to_string term_to_string "," termlist)^")"*)

  | Base(Var(id)) | Base(Const(id))
  | Prime(Var(id)) | Prime(Const(id)) ->
      Globals.string_id#get_name id
  | Hash(_) ->
      prerr_endline "Interface.term_to_string: Impossible error: invalid hash type/term"; 
      raise Not_found
;;

let print_term (t : Types.atoms_type Types.term_structure) : unit =
   prout_string (term_to_string t)
;;
let prerr_term (t : Types.atoms_type Types.term_structure) : unit =
   prerr_string (term_to_string t)
;;



(*
   This function is similar to the previous one except that it use HLPSL terms and not IF terms.
  \begin{center}
  type:
   [val term_value_to_string : int -> Types.atoms_type Types.term_structure -> string]
  \end{center}
*)
let rec term_value_to_string (session_id : int) : Types.atoms_type Types.term_structure -> string =
  (* This sub-fonction prints the IF values of atoms.*)
  let atom_value_to_string hlpsl_id = 
    let gid = global_var_id#get_id_of session_id hlpsl_id in
    let map_id = memory_map#get_map gid in
    let if_value = value_table#get_value map_id in
    ground_if_value_to_string if_value
  in function
      Pair(term1,term2) ->
      	"pair("^(term_value_to_string session_id term1)^","
      	^(term_value_to_string session_id term2)^")"
    | Scrypt(term1,term2) ->
      	"scrypt("^(term_value_to_string session_id term2)^","
      	^(term_value_to_string session_id term1)^")"
    | Crypt(term1,term2) ->
      	"crypt("^(term_value_to_string session_id term2)^","
      	^(term_value_to_string session_id term1)^")"
    | Inv(term) ->
      	"inv("^(term_value_to_string session_id term)^")"
    | Function(term1,termlist) ->
      	(term_value_to_string session_id term1)^"("
      	^(list_to_string (term_value_to_string session_id) "," termlist)^")"
    | Cons(term1,term2) ->
      	"cons("^(term_value_to_string session_id term1)^","
      	^(term_value_to_string session_id term2)^")"
    | Delete(term1,term2) ->
      	"delete("^(term_value_to_string session_id term1)^","
      	^(term_value_to_string session_id term2)^")"
    | Set(termlist) ->
      	"set("
      	^(list_to_string (term_value_to_string session_id) "," termlist)^")"
(*BUG??? 15/03/2005
      	list2_to_string (term_value_to_string session_id) "," termlist *)
(*    | Lst(termlist) ->
      	"list("
      	^(list_to_string (term_value_to_string session_id) "," termlist)^")"*)
    | Prime(Var(id))
    | Prime(Const(id)) ->
      	(Globals.string_id#get_name id)^"'"
    | Base(Var(id)) ->
      	atom_value_to_string id
    | Base(Const(id)) ->
      	string_id#get_name id;
    | Hash(_) ->
	prerr_endline "Interface.term_value_to_string: Impossible error: invalid hash type/term"; 
	raise Not_found
;;


(*

   \subsection{Printing \textsf{Intermediate Format} facts}

   [print_if_term] prints IF facts. It also generates output for facts signatures.
  \begin{center}
   type: [val if_term_to_string : string -> Types.if_term -> string]\\
   type: [val print_if_term : string -> Types.if_term -> unit]
   type: [val prerr_if_term : string -> Types.if_term -> unit]
  \end{center}
*)
let rec if_term_to_string (func : string) : Types.if_term -> string =

  let type_to_string = function
      Base(Var(i))
    | Base(Const(i)) ->
	term_type_to_string (Globals.type_table#get_type i)
    | Set(_) ->
	"set"
    | _ ->
	"message";
  in function

      State(name,session_id,l) ->
      	if (func = "print_term_type") then
	  if not(List.mem name !roles_printed) then
	    (roles_printed := !roles_printed @ [name];
	     ("state_"^name^" : "^(list_to_string type_to_string " * " l)^" -> fact\n"))
	  else ""
      	else 
	  (
	   (if func = "print_term_value" then 
	     list_to_string (term_value_to_string session_id) "," l
	   else
	     list_to_string term_to_string "," l)
	   )

    | Iknows t -> 
      	if not (func = "print_term_type") then
	  ((term_to_string t))
	else ""

    | Ifcontains (t1,t2) -> 
      	if not (func = "print_term_type") then
	  ("contains("^(term_to_string t1)
	   ^","^(term_to_string t2)^")")
	else ""

    | Ifdelete (t1,t2) -> 
      	if not (func = "print_term_type") then
	  ("contains("^(term_to_string t1)
	   ^","^(term_to_string t2)^")")
	else ""

    | Ifequal (t1,t2) -> 
      	if not (func = "print_term_type") then
	  ("equal("^(term_to_string t1)
	   ^","^(term_to_string t2)^")")
	else ""

    | Ifleq (t1,t2) -> 
      	if not (func = "print_term_type") then
	  ("leq("^(term_to_string t1)
	   ^","^(term_to_string t2)^")")
	else ""

    | Ifnot t ->
 	"not("^(if_term_to_string func t)^")"

(*    | Secret(t1,t2) -> 
     	if not (func = "print_term_type") then
	  ("secret("^(term_to_string t1)^","
	   ^(term_to_string t2)^")")
	else ""*)

(*    | Witness(authenticator, authenticated, const_term, var_term) ->
     	if not (func = "print_term_type") then
	  let pl_authenticator = find_player_in_role authenticator in
	  let pl_authenticated = find_player_in_role authenticated in
	  ("witness("
	   ^(Globals.string_id#get_name pl_authenticator)^","
	   ^(Globals.string_id#get_name pl_authenticated)^","
	   ^(term_to_string const_term)^","
	   ^(term_to_string var_term)
	   ^")")
	else ""*)
(*Event: *)
    | Unknown(t,lt) ->
      	(match t with
	  Base(Const(id)) ->
	    let nid = Globals.string_id#get_name id in
	    if (nid = "request") or (nid = "wrequest") then
	      ""(*(nid^"("^(list_to_string term_to_string "," lt)^",SID)")*)
	    else
	      (nid^"("^(list_to_string term_to_string "," lt)^")")
	| _ ->
	    (term_to_string t)^"("
 	    ^(list_to_string term_to_string "," lt)^")")

    | _ ->
 	Error_handler.displayWarning 5 "print_if_term"; ""
;;

let print_if_term (func : string) (t : Types.if_term) : unit =
  prout_string (if_term_to_string func t)
;;
let prerr_if_term (func : string) (t : Types.if_term) : unit =
  prerr_string (if_term_to_string func t)
;;


let rec ltl_formula_to_string : Types.ltl_type -> string = function
    Pred(Ifnot(p)) -> "~ "^(ltl_formula_to_string (Pred(p)))
  | Pred(p) -> if_term_to_string "" p
  | Op1(f) -> "[] "^(ltl_formula_to_string f)
  | Op2(f) -> "<-> "^(ltl_formula_to_string f)
  | Op3(f) -> "(-) "^(ltl_formula_to_string f)
  | Op4(f) -> "[-] "^(ltl_formula_to_string f)
  | Op5(f) -> "~ "^(ltl_formula_to_string f)
  | Conj(f1,f2) -> "("^(ltl_formula_to_string f1)^" /\\ "^(ltl_formula_to_string f2)^")"
  | Dij(f1,f2) -> "("^(ltl_formula_to_string f1)^" \\/ "^(ltl_formula_to_string f2)^")"
  | Impl(f1,f2) -> "("^(ltl_formula_to_string f1)^" => "^(ltl_formula_to_string f2)^")"
;;


(*
  \subsection{Translation from [atom_type] to [string]}
*)

(*
  \begin{center}
  type:
   [val atoms_type_to_string : Types.basic_type Types.term_structure -> string]
  \end{center}
*)
let rec atoms_type_to_string (parentheses : bool) (a_term : Types.basic_type Types.term_structure) : string =
    match a_term with
      Base(term) | Prime(term) ->
	(match term with
	  Agent -> "agent"
	| Public_key -> "public_key"
	| Symmetric_key -> "symmetric_key"
	| Text -> "text"
	| Message -> "message"
	| Protocol_id -> "protocol_id"
	| Channel -> "channel"
	| Channel_dy -> "channel_dy"
	| Channel_ota -> "channel_ota"
	| Bool -> "bool"
	| Hash_func -> "hash_func"
	| Nat  -> "nat"
(*	| Pointer -> "set"*)
	| Enum(list) ->
	    "{"
	    ^(List.fold_left
		(fun s num ->
	      	  if not (num = List.hd list) then
	      	    s^","^(string_id#get_name num)
	      	  else
		    s^(string_id#get_name num))
		"" list)
	    ^"}")
    | Inv t -> "inv("^(atoms_type_to_string false t)^")";
    | Hash t -> "apply(hash_func,"^(atoms_type_to_string false t)^")";
    | Set [] -> "set"
    | Set lt -> "set("^(atoms_type_to_string false (List.hd lt))^")"
(*    | Lst [] -> "list"
    | Lst lt -> "list("^(atoms_type_to_string false (List.hd lt))^")"*)
    | Pair(t1,t2) ->
	"pair("^(atoms_type_to_string false t1)^","^(atoms_type_to_string false t2)^")"
    | Crypt(t1,t2) -> "crypt("^(atoms_type_to_string false t2)^","^(atoms_type_to_string false t1)^")"
    | Scrypt(t1,t2) -> "scrypt("^(atoms_type_to_string false t2)^","^(atoms_type_to_string false t1)^")"
    | Function(t,lt) ->
	let string_of_args =
	  match lt with
	    [] -> "unit"
	  | [t1] -> atoms_type_to_string true t1
	  | t1::l ->
	      List.fold_left
		(fun s t ->
		  s^" * "^(atoms_type_to_string true t))
		(atoms_type_to_string true t1) l
	in
	if parentheses then
	  "("^string_of_args^" -> "^(atoms_type_to_string true t)^")"
	else
	  string_of_args^" -> "^(atoms_type_to_string true t)
    | _ -> "message"
;;

(*
  \begin{center}
  type:
   [val global_id_to_type_string : int -> string]
  \end{center}
*)
let global_id_to_type_string (gid : int) :string =
  atoms_type_to_string true
    (Globals.type_table#get_type
       (Globals.global_var_id#get_name_of gid))
;;


(*
  \section{Printing the output file}
*)


(*

  \subsection{Printing the prelude}
 
   The prelude from the compiler's point of view is a file
   [prelude.if] containing static data, containing some
   meta-information about the displayed IF sections.  It also contains
   the intruder section, describing its main actions.  This file is
   independant, and therefore not printed by the compiler; it has to
   be considered by the backends.
*)

(*

  \subsection{Printing the types section}

*)


(*
  \begin{center}
  type:
   [val print_section_types : unit -> unit]
  \end{center}
*)


let print_section_types (() : unit) : unit =
  (* Computation of the list of ids of basic roles *)
  let role_ids = 0::(List.map snd !Globals.player_to_role) in
  (*
     The key  in the [identifiers_type_table] is a string representing 
     a type of an identifier as it will be printed. The [int] is the [string_id]
     corresponding to the identifier.
   *)
  let identifiers_type_table = (Hashtbl.create 15 : (string,int) Hashtbl.t) in
  let identifiers_type_table_key_list = ref ([]: string list) in
  Hashtbl.iter
    (fun x y ->
      (* If x is a variable, it has to be declared in a basic role *)
	if (not_excluded_type y)
	   && (List.mem (Globals.string_id#get_role_id x) role_ids) then
	  (let y_string = (atoms_type_to_string false y) in
	   Hashtbl.add identifiers_type_table y_string x;
	   identifiers_type_table_key_list := 
	     Utils.insert_with_no_copy y_string !identifiers_type_table_key_list))
    Globals.type_table#get_tab;

  List.iter
    (fun a ->
      let list_of_identifiers_of_type_a = 
	Utils.remove_duplicate
	  (List.map 
	     (Globals.string_id#get_name; ) 
	     (Hashtbl.find_all identifiers_type_table a))
      in
      List.iter (fun t -> (add_type (replace "dummy_" "" t) [a])) list_of_identifiers_of_type_a;)
    !identifiers_type_table_key_list
;;   

(* Verify if the LTL Formula is convertible. To be convertible a
   formula can not cantain other temporal operator than []
   val is_convertible_property : ltl_type -> bool *)
let rec is_convertible_property : Types.ltl_type -> bool = function
  | Pred(str) -> true

  | Op1(f) | Op2(f) | Op3(f) | Op4(f) -> false

  (* ~f *)
  | Op5(f)    -> is_convertible_property f

  (* f1 /\ f2 *)
  | Conj(f1,f2) -> (is_convertible_property f1) && (is_convertible_property f2)

  (* f1 \/ f2 *)
  | Dij(f1,f2) -> (is_convertible_property f1) && (is_convertible_property f2)

  (* f1 -> f2 *) 
  | Impl(f1,f2) -> (is_convertible_property f1) && (is_convertible_property f2)
;;


(* Convert an LTL formula to a Disjunctive Normal Form (DNF) 
   val ltl_to_dnf : bool -> ltl_type -> ltl_type *)
let ltl_to_neg_dnf (f : Types.ltl_type) : Types.if_term list list =
  let lists_from_dnf f =

    let rec list_from_dij = function
      | Dij(f1,f2) -> (list_from_dij f1) @ (list_from_dij f2)
      | x -> [x]
    in
    let rec list_from_conj = function
      | Pred(str) -> [str]
      | Conj(f1,f2) -> (list_from_conj f1) @ (list_from_conj f2)
      | x -> failwith "list_from_conj: impossible case"
    in

    List.map list_from_conj (list_from_dij f)
  in

  let rec ltl_to_dnf neg = function
    | Pred(Ifnot(str)) ->
	if neg then
	  Pred(str)
	else
	  Pred(Ifnot(str))

    | Pred(str) ->
	if neg then
	  Pred(Ifnot(str))
	else
	  Pred(str)

	    (* ~f *)
    | Op5(f)    -> 
	ltl_to_dnf (not neg) f

	  (* f1 /\ f2 *)
    | Conj(f1,f2) ->
	if neg then
	  ltl_to_dnf false (Dij(Op5(f1),Op5(f2)))
	else
	  let f1' = ltl_to_dnf false f1
	  and f2' = ltl_to_dnf false f2
	  in
	  (match f1' with
	    Dij(f1a,f1b) ->
	      Dij(ltl_to_dnf false (Conj(f1a,f2')),
		  ltl_to_dnf false (Conj(f1b,f2')))
	  | _ ->
	      (match f2' with
		Dij(f2a,f2b) ->
		  Dij(ltl_to_dnf false (Conj(f1',f2a)),
		      ltl_to_dnf false (Conj(f1',f2b)))
	      | _ ->
		  Conj(f1',f2')))

	    (* f1 \/ f2 *)
    | Dij(f1,f2) ->
	if neg then
	  ltl_to_dnf false (Conj(Op5(f1),Op5(f2)))
	else
	  let f1' = ltl_to_dnf false f1
	  and f2' = ltl_to_dnf false f2
	  in
	  Dij(f1',f2')

	    (* f1 => f2 *)
    | Impl(f1,f2) ->
	ltl_to_dnf neg (Dij(Op5(f1),f2))

    | x -> x (* impossible case *)
  in

  lists_from_dnf (ltl_to_dnf true f)
;;


(*
   \subsection{Printing the goals section}
  \begin{center}
  type:
   [val print_section_goals : Types.goals_type list -> unit]
  \end{center}
*)
let print_section_goals (print_previous_sections : 	(Types.atoms_type Types.term_structure list ->
        						Types.atoms_type Types.term_structure list ->
        						Types.atoms_type Types.term_structure list -> 'a)) 
			(goals : Types.goals_type list)  : unit =
  (* filename: name of the output file, in case of goals splitting *)
  let filename = ref "" in
  (* secrecy_info: list of ID of informations used for secrecy *)
  let secrecy_info = ref [] in
  (* auth_info: list of ID of informations used for authentication *)
  let auth_info = ref [] in
  (* wauth_info: list of ID of informations used for weak authentication *)
  let wauth_info = ref [] in
  (* secrecy: number of secrecy goals *)
  let secrecy = ref 0 in
  (* auth: number of authentication goals *)
  let auth = ref 0 in
  (* wauth: number of weak authentication goals *)
  let wauth = ref 0 in
  (* ltl: number of LTL goals *)
  let ltl = ref 0 in

  let print_secret_attack_state xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state secrecy_of_"^sk^" (MGoal,ASGoal) :=");
    prout_endline " iknows(MGoal).";
    prout_endline (" secret(MGoal,"^sk^",ASGoal) &");
    prout_endline " not(contains(i,ASGoal))";
    prout_newline()
  in
  let print_secret_property xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-secrecy-"^sk^".if";
       prerr_endline ("%% IF output in "^(!filename));
       Globals.tmp_channel := open_out !filename;
      (*prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");*)
       print_previous_sections [Base(Const(xid))] [] [];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property secrecy_of_"^sk^" (MGoal,ASGoal) :=");
    prout_endline (" [] ((secret(MGoal,"^sk^",ASGoal) /\\ iknows(MGoal))");
    prout_endline "     => contains(i,ASGoal))";
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_secret_attack_state xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_authentication_attack_state xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state authentication_on_"^sk^" (A1Goal,A2Goal,MGoal,SID) :=");
    prout_endline (" request(A1Goal,A2Goal,"^sk^",MGoal,SID) &");
    prout_endline (" not(witness(A2Goal,A1Goal,"^sk^",MGoal)) &");
    prout_endline (" not(equal(A2Goal,i))");
    prout_endline ("attack_state replay_protection_on_"^sk^" (A2Goal,A1Goal,MGoal,SID1,SID2) :=");
    prout_endline (" request(A1Goal,A2Goal,"^sk^",MGoal,SID1).");
    prout_endline (" request(A1Goal,A2Goal,"^sk^",MGoal,SID2) &");
    prout_endline (" not(equal(SID1,SID2)) &");
    prout_endline (" not(equal(A2Goal,i))");
    prout_newline()
  in
  let print_authentication_property xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-auth-"^sk^".if";
       (*prerr_endline ("%% IF output in "^(!filename));*)
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [] [Base(Const(xid))] [];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property authentication_on_"^sk^" (A1Goal,A2Goal,MGoal,SID,SID1,SID2) :=");
    prout_endline (" [] (((request(A1Goal,A2Goal,"^sk^",MGoal,SID)");
    prout_endline ("       /\\ ~ equal(A2Goal,i))");
    prout_endline ("      => witness(A2Goal,A1Goal,"^sk^",MGoal))");
    prout_endline ("     /\\ ((request(A1Goal,A2Goal,"^sk^",MGoal,SID1)");
    prout_endline ("           /\\ request(A1Goal,A2Goal,"^sk^",MGoal,SID2)");
    prout_endline ("           /\\ ~ equal(A2Goal,i))");
    prout_endline "          => equal(SID1,SID2)))";
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_authentication_attack_state xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_weak_authentication_attack_state xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state weak_authentication_on_"^sk^" (A1Goal,A2Goal,MGoal,SID) :=");
    prout_endline (" wrequest(A1Goal,A2Goal,"^sk^",MGoal,SID) &");
    prout_endline (" not(witness(A2Goal,A1Goal,"^sk^",MGoal)) &");
    prout_endline (" not(equal(A2Goal,i))");
    prout_newline()
  in
  let print_weak_authentication_property xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-wauth-"^sk^".if";
       (*prerr_endline ("%% IF output in "^(!filename));*)
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [] [] [Base(Const(xid))];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property weak_authentication_on_"^sk^" (A1Goal,A2Goal,MGoal,SID) :=");
    prout_endline (" [] ((wrequest(A1Goal,A2Goal,"^sk^",MGoal,SID)");
    prout_endline ("      /\\ ~ equal(A2Goal,i))");
    prout_endline ("     => witness(A2Goal,A1Goal,"^sk^",MGoal))");
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_weak_authentication_attack_state xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_auth_attack_state s1 s2 xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state authentication_on_"^sk^" ("
		   ^s2^","^s1^",MGoal,SID) :=");
    prout_endline (" request("^s1^","^s2^","
		   ^sk^",MGoal,SID) &");
    prout_endline (" not(witness("^s2^","^s1^","
		   ^sk^",MGoal)) &");
    prout_endline (" not(equal("^s2^",i))");
    prout_endline ("attack_state replay_protection_on_"^sk^" ("
		   ^s2^","^s1^",MGoal,SID1,SID2) :=");
    prout_endline (" request("^s1^","^s2^","
		   ^sk^",MGoal,SID1).");
    prout_endline (" request("^s1^","^s2^","
		   ^sk^",MGoal,SID2) &");
    prout_endline (" not(equal(SID1,SID2)) &");
    prout_endline (" not(equal("^s2^",i))");
    prout_newline()
  in
  let print_auth_property s1 s2 xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-auth-"^(string_of_int !auth)^".if";
       (*prerr_endline ("%% IF output in "^(!filename));*)
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [] [Base(Const(xid))] [];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property authentication_on_"^sk^" ("
		   ^s2^","^s1^",MGoal,SID,SID1,SID2) :=");
    prout_endline (" [] (((request("^s1^","^s2^","^sk^",MGoal,SID) /\\ ~ equal(A2Goal,i))");
    prout_endline ("      => witness("^s2^","^s1^","^sk^",MGoal))");
    prout_endline ("     /\\ ((request("^s1^","^s2^","^sk^",MGoal,SID1)");
    prout_endline ("           /\\ request("^s1^","^s2^","^sk^",MGoal,SID2)");
    prout_endline ("           /\\ ~ equal(A2Goal,i))");
    prout_endline "           => equal(SID1,SID2)))";
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_auth_attack_state s1 s2 xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_weak_auth_attack_state s1 s2 xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state weak_authentication_on_"^sk^" ("
		   ^s2^","^s1^",MGoal,SID) :=");
    prout_endline (" wrequest("^s1^","^s2^","
		   ^sk^",MGoal,SID) &");
    prout_endline (" not(witness("^s2^","^s1^","
		   ^sk^",MGoal)) &");
    prout_endline (" not(equal("^s2^",i))");
    prout_newline()
  in
  let print_weak_auth_property s1 s2 xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-wauth-"^(string_of_int !wauth)^".if";
       (*prerr_endline ("%% IF output in "^(!filename));*)
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [] [] [Base(Const(xid))];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property weak_authentication_on_"^sk^" ("^s1^","^s2^",MGoal,SID) :=");
    prout_endline (" [] ((wrequest("^s1^","^s2^","^sk^",MGoal,SID)");
    prout_endline ("      /\\ ~ equal(A2Goal,i))");
    prout_endline ("     => witness("^s2^","^s1^","^sk^",MGoal))");
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_weak_auth_attack_state s1 s2 xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_ltl_attack_states p lvc =
    let las = ltl_to_neg_dnf p in
    let ltl2 = ref 0 in
    List.iter
      (fun s ->
	incr ltl2;
	prout_endline ("attack_state ltl_"^(string_of_int !ltl)^"_"^(string_of_int !ltl2)^" ("
		       ^(list_to_string term_to_string "," lvc)^") :=");
	prout_string " ";
	let (s1,s2) =
	  List.partition
	    (function
		Ifnot(_)
	      | Ifequal(_)
	      | Ifleq(_) -> false
	      | _ -> true)
	    s
	in
	print_list (if_term_to_string "") ".\n " s1;
	if s2 <> [] then
	  (if s1 <> [] then prout_string " &\n ";
	   print_list (if_term_to_string "") " &\n " s2);
	prout_newline())
      las;
    prout_newline()
  in
  let print_ltl_property p lvc =
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-ltl-"^(string_of_int !ltl)^".if";
       (*prerr_endline ("%% IF output in "^(!filename));*)
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
(*** WARNING: all goal facts are printed!  to be changed when real LTL formulas supported? ***)
       print_previous_sections [] [] [];
       prout_endline "section properties:"; prout_newline ());
    prout_string "property ltl_";
    prout_int !ltl;
    prout_endline (" ("^(list_to_string term_to_string "," lvc)^") :=");
    prout_endline (" "^(ltl_formula_to_string p));
    prout_newline();
    if !Globals.flag_split
	&& (is_convertible_property p) then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_ltl_attack_states p lvc;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  if not (!Globals.flag_split) then
    (let secrecy_ids = ref [] in
    let auth_ids = ref [] in
    let wauth_ids = ref [] in
    List.iter
      (function
	  Secrecy(lt) -> secrecy_ids:= !secrecy_ids @ (List.map (fun x -> Base(Const x)) lt)
	| Auth(_,_,lt)
	| Authentication(lt) -> auth_ids:= !auth_ids @ (List.map (fun x -> Base (Const x)) lt)
	| Weak_auth(_,_,lt)
	| Weak_authentication(lt) ->  wauth_ids:= !wauth_ids @ (List.map (fun x -> Base (Const x)) lt)
	| LTL_goal(_) -> ())
      goals;
    print_previous_sections !secrecy_ids !auth_ids !wauth_ids;
    prout_endline "section properties:"; prout_newline ());

  (* first pass: for printing properties, and if splitting is required, attack states *)
  List.iter
    (function
	Secrecy(lt) ->
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !secrecy_info)) then
		(secrecy_info := xid::!secrecy_info;
		 incr secrecy;
		 print_secret_property xid))
	    lt

      | Auth(r1,r2,lt) ->
	  let name_r1 =
	    Globals.string_id#get_name (find_player_in_role r1)
	  and name_r2 =
	    Globals.string_id#get_name (find_player_in_role r2)
	  in
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !auth_info)) then
		(auth_info := xid::!auth_info;
		 incr auth;
		 print_auth_property name_r1 name_r2 xid))
	    lt
      | Authentication(lt) ->
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !auth_info)) then
		(auth_info := xid::!auth_info;
		 incr auth;
		 print_authentication_property xid))
	    lt

      | Weak_auth(r1,r2,lt) ->
	  let name_r1 =
	    Globals.string_id#get_name (find_player_in_role r1)
	  and name_r2 =
	    Globals.string_id#get_name (find_player_in_role r2)
	  in
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !wauth_info)) then
		(wauth_info := xid::!wauth_info;
		 incr wauth;
		 print_weak_auth_property name_r1 name_r2 xid))
	    lt
      | Weak_authentication(lt) ->
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !wauth_info)) then
		(wauth_info := xid::!wauth_info;
		 incr wauth;
		 print_weak_authentication_property xid))
	    lt

      | LTL_goal(s,lvc) -> 
	  incr ltl;
	  print_ltl_property s (List.filter (function Base(Var(_)) -> true | _ -> false) lvc))
    goals;

  (* second pass: for printing attack states if no splitting required *)
  if (not !Globals.flag_split)
      && (List.exists
	    (function
		LTL_goal(Op1(s),_) -> is_convertible_property s
	      | _ -> true)
	    goals) then
    (prout_newline ();
     prout_endline "section attack_states:";
     prout_newline ();
     secrecy_info := [];
     auth_info := [];
     wauth_info := [];
     ltl := 0;
     List.iter
       (function
	   Secrecy(lt) ->
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !secrecy_info) then
		   (secrecy_info := xid::!secrecy_info;
		    incr secrecy;
		    print_secret_attack_state xid))
	       lt

	 | Auth(r1,r2,lt) ->
	     let name_r1 =
	       Globals.string_id#get_name (find_player_in_role r1)
	     and name_r2 =
	       Globals.string_id#get_name (find_player_in_role r2)
	     in
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !auth_info) then
		   (auth_info := xid::!auth_info;
		    incr auth;
		    print_auth_attack_state name_r1 name_r2 xid))
	       lt
	 | Authentication(lt) ->
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !auth_info) then
		   (auth_info := xid::!auth_info;
		    incr auth;
		    print_authentication_attack_state xid))
	       lt

	 | Weak_auth(r1,r2,lt) ->
	     let name_r1 =
	       Globals.string_id#get_name (find_player_in_role r1)
	     and name_r2 =
	       Globals.string_id#get_name (find_player_in_role r2)
	     in
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !wauth_info) then
		   (wauth_info := xid::!wauth_info;
		    incr wauth;
		    print_weak_auth_attack_state name_r1 name_r2 xid))
	       lt
	 | Weak_authentication(lt) ->
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !wauth_info) then
		   (wauth_info := xid::!wauth_info;
		    incr wauth;
		    print_weak_authentication_attack_state xid))
	       lt

	 | LTL_goal(Op1(s),lvc) -> 
	     incr ltl;
	     if is_convertible_property s then
	       (print_ltl_attack_states s
		  (List.filter (function Base(Var(_)) -> true | _ -> false) lvc))

	 | _ -> ())
       goals)
;;

(***** ***** ***** ***** *****)
	 	
(*
  \begin{center}
  type:
   [val print_section_signatureA : role_instance list -> unit]
  \end{center}
*)
(**
record the signature of the roles
@return unit
*)
let print_section_signatureA (role_instances : role_instance list) : unit = 
  (*
     The key in the [role_type_table] is an integer representing the 
     key in the [Globals.string_id] table of the name
     of a role.  The second argument is the string representing 
     the type of the role. The [ -> Fact] part is missing.
   *) 
  let role_type_table = (Hashtbl.create 10 : (int,(string list)) Hashtbl.t) in
  let role_type_table_key_list = ref ([]: int list) in 
  let init_state_list = 
    List.map 
      (fun role_instance ->
(*	(role_instance#get_name, (List.hd role_instance#get_initial_state))*)
	(role_instance#get_name, role_instance#get_initial_state);)
      role_instances 
  in
  List.iter
    (fun (a,lb) ->
      List.iter
	(function
	    (*
	       We are only interested in the list of actual arguments since we already
	       know the role identity.
	     *)
	    State(_,_,l) ->
	      let string_from_type =
		match l with
		  [] -> []
		| Base(Var(head))::tail ->
		    List.fold_left 
		      (fun already_done -> function
			  Base(Var(elt)) -> 
			    (already_done@[(global_id_to_type_string elt)]) 
			| _ -> raise Not_found)
		      [(global_id_to_type_string head)] tail
		| _ -> raise Not_found
	      in
	      Hashtbl.add role_type_table a string_from_type;
	      role_type_table_key_list := Utils.insert_with_no_copy a !role_type_table_key_list
	  | _ -> ())
	lb)
    init_state_list;
    
    List.iter
    	(fun a -> 
	(add_type (Globals.string_id#get_name a) ("nat"::(Hashtbl.find role_type_table a)))
	)
    !role_type_table_key_list
 ;;
 
 (** contains the initial states of all the agent. Each pair is (name_of _the_agent, initial_state) *)
 let initial_states = (Hashtbl.create 600 : (string, string) Hashtbl.t);;
 
 (**
 Record an initial state "stae" of the agent "name" in the hash table initial_states
 @param name is an agent
 @param state is an initial stae of the agent "name"
 @return unit
 *)
 let add_state (name : string) (state : string) : unit = 
	Hashtbl.add initial_states name (replace "dummy_" "" state);; 

(** return the initial state of the agent "name"
 @param name is an agent
 @return the initial state of the agent "name"
 *)
 let get_initial_state (name : string) : string = 
	if (Hashtbl.find_all initial_states name=[])
	then ""
	else Hashtbl.find initial_states name ;; 

let list_to_stringA (f :  ('a -> unit)) :  'a list -> unit =
  let rec list_to_string_rec = function
      [] -> ()
    | [h] -> f h
    | h::t -> ((f h) ; (list_to_string_rec t))
  in
  list_to_string_rec
;;

(*
  \begin{center}
  type:
   [val print_initial_state : Types.if_term list -> unit]
  \end{center}
 *)
 (**
 Record the initial state of the roles
 @return init
 *)
let print_initial_stateA : 'a list -> unit =
  let one_argument_to_string arg = match arg with
      Base(Var(id)) -> if_value_to_string atom_ground_value_to_string (If_address(id))
    | _ -> "Interface.prout_initial_state: Not a variable"
  in
  let one_state_to_string = function
      State(role_string,id,term_list) ->
	(add_state (get_name_in_session role_string id) (list_to_string one_argument_to_string "," term_list))
    | _ ->
	failwith "Interface.print_initial_state: Warning: Predicate not handled!"; 
  in
    list_to_stringA one_state_to_string 
;;


  
 (*
  \begin{center}
  type:
   [val print_section_init : role_instance list -> unit]
  \end{center}
*)
(**
record the initial states, the initial knowledge of the intruder
@return unit
*)
let print_section_init (role_instances : role_instance list) : unit= 
  let init_state =
    List.flatten
      (List.map 
	 (fun role_instance ->
	   role_instance#get_initial_state)
	 role_instances)
  in
  if (init_state <> []) then
    (print_initial_stateA init_state);
    
        
  List.iter
    (fun x ->
    	let k = (ground_if_value_to_string x) in
	if(not(List.mem k !intruder_knowledge)) then intruder_knowledge :=  k::(!intruder_knowledge))
    !Globals.i_knowledge;

  contains_term_list := (Utils.remove_duplicate !contains_term_list);
  List.iter
    (fun (a,b) ->
	prerr_string (ground_if_value_to_string a);	
	prerr_string (ground_if_value_to_string b);
    add_in_set "x" (ground_if_value_to_string a) "x" (ground_if_value_to_string b))
    !contains_term_list;
  contains_term_list := [];
  
    
;;

(* ---------------------------------- *)

(*
   Printing the IF states using formatting function "func".
  \begin{center}
  type:
   [val print_if_states_listA : string -> Types.if_term list -> string]
  \end{center}
*) 
let print_if_states_listA (func : string) (l : Types.if_term list) : string * string list =
  let tmp= ref "" in
  let res = ref [] in
  let rec print_if_states_list_recA = function
      [] -> Error_handler.displayWarning 5 "print_if_states"
    | [h] ->
    	(tmp := (if_term_to_string func h);
    	if((!tmp)<>"") then  (res:= (!tmp)::(!res)))
    | h::t ->
    	(tmp:= (if_term_to_string func h);
        if((!tmp)<>"") then (res := (!tmp)::(!res))) ;
      	print_if_states_list_recA t
  in
  print_if_states_list_recA l;
  match (List.rev (!res)) with  [state] -> (state, [""])
  				|(state::iknows) -> (state, iknows)
  				|_ -> failwith "Bad state"
;;

 (* *** Print rules *** *)
 
 (** print the transitions of the agents *)
 let multi_t = (ref [] : (string * int * (('a * 'b * 'c) list)) list Pervasives.ref);;
 let test_multi_rcv name_in_session state mess1 mess2 =
	let rec aux2 actionsL bool1 bool2 =
		match actionsL with [] -> false
		| (initial,_,final)::ls -> 	let tmp_bool1 = if(bool1) then bool1 else (((findState final)=state) && ((findIknows initial)=mess1)) in
						let tmp_bool2 = if(bool2) then bool2 else (((findState final)=state) && ((findIknows initial)=mess2)) in
						if (bool1 && bool2)
						then true
						else aux2 ls tmp_bool1 tmp_bool2
	in
	let rec aux l =
		match l with [] -> false
		| (name,id,actionsL)::ls -> 	if((get_name_in_session name id) = name_in_session && (aux2 actionsL false false))
						then true
						else aux ls; 
	in aux (!multi_t)		
 ;;
 let multi_transition (name : string) (id : int) (initial : string * string list) (cond : string list) (exists_var_list : string list) (final : string * string list) :  
 			((string * string) *
        			((string list * string list) * (string list * string list)) *
        		(string * string)) list =
 	let var_new s =
		let res = ref s in
		List.iter (fun v -> res := (replace (","^v^",") (","^v^"_new,") (!res))) exists_var_list;
		!res
	in
	let var_new_iknows s =
		let res = ref s in
		List.iter 
		(fun v ->
			if ((!res)=v) then res := ((!res)^"_new")
			else(
				res := (replace (","^v^")") (","^v^"_new)") (!res));
				res := (replace ("("^v^",") ("("^v^"_new,") (!res));
				res := (replace (","^v^",") (","^v^"_new,") (!res))
			)
		) 
		exists_var_list;
		!res
	in
	let replace_head head l = 
		if(l<>[]) then (head^"("^(list2string "," (cdr l))^")")
		else head
	in
	let rec find_iknows_contains_left l_snd l_other iknows test affectation =
		match l_snd with [] -> (List.rev iknows, 
					List.rev test, 
					List.rev affectation)
			|(x::ls) -> 	let d = decompose x in
					if(d<>[])
					then(	
						let dd = decompose (cadr d) in
						if((car d) = "iknows" ) 
						then find_iknows_contains_left ls l_other (x::iknows) test affectation
						else if((car d) = "contains" && not(List.mem x l_other)) 
						then  find_iknows_contains_left ls l_other iknows (x::test) ((replace_head "delete" d)::affectation)
						else if((car d) = "contains" && (List.mem x l_other) && not(List.mem x test)) 
						then  find_iknows_contains_left ls l_other iknows (x::test) affectation
						else if ((car d)="not" && (car dd)="contains" && not(List.mem (cadr d) l_other))
						then find_iknows_contains_left ls l_other iknows test ((replace_head "delete" dd)::affectation)
						else if((car d) = "leq" || (car d) = "equal" || (car d) = "idequal" || (car d) = "not") 
						then find_iknows_contains_left ls l_other iknows (x::test) affectation
						else if((car d)<>"secret" && (car d)<>"witness" && (car d)<>"request")
						then find_iknows_contains_left ls l_other (x::iknows) test affectation
						else  find_iknows_contains_left ls l_other iknows test affectation
					)
					else find_iknows_contains_left ls l_other (x::iknows) test affectation;
	in
	let rec find_iknows_contains_right l_snd l_other iknows test affectation =
		match l_snd with [] -> (List.rev iknows, 
					List.rev test, 
					List.rev affectation)
			|(x::ls) -> 	let d = decompose x in
					if(d<>[])
					then(
						if((car d) = "iknows" ) 
						then find_iknows_contains_right ls l_other (x::iknows) test affectation
						else if((car d) = "contains" && not(List.mem x l_other)) 
						then  find_iknows_contains_right ls l_other iknows test ((replace_head "cons" d)::affectation)
						else if((car d) = "contains" && (List.mem x l_other) && not(List.mem x test)) 
						then  find_iknows_contains_right ls l_other iknows (x::test) affectation
						else if((car d) = "leq" || (car d) = "equal" || (car d) = "idequal" || (car d) = "not") 
						then find_iknows_contains_right ls l_other iknows (x::test) affectation
						else if((car d)<>"secret" && (car d)<>"witness" && (car d)<>"request")
						then find_iknows_contains_right ls l_other (x::iknows) test affectation
						else  find_iknows_contains_right ls l_other iknows test affectation
					)
					else find_iknows_contains_right ls l_other (x::iknows) test affectation;
	in		
	let fst_final_new = var_new (fst final) in
	let (iknows_initial, op_test_initial, op_affectation_initial) = find_iknows_contains_left ((snd initial)@cond) (snd final) [] [] [] in
	let (iknows_final, op_test_final, op_affectation_final) = find_iknows_contains_right (snd final) ((snd initial)@cond) [] [] [] in
	let iknows_final0 = if(iknows_final<>[]) then iknows_final else [""] in
	let iknows_initial0 = if(iknows_initial<>[]) then iknows_initial else [""] in
	let res = List.flatten (List.map 
	(fun x -> (List.map (fun y -> ((fst(initial), x),((op_test_initial,op_affectation_initial),(op_test_final,op_affectation_final)),(fst_final_new, (var_new_iknows y) ))) iknows_final0)) 
	iknows_initial0) in
	multi_t := (name, id, res)::(!multi_t); 
	res
;;

	(*****************)
let print_if_states_list func l =
  let rec print_if_states_list_rec = function
      [] -> Error_handler.displayWarning 5 "print_if_states"
    | [h] ->
      	if (func <> "print_term_type") then
	  prout_string " ";
	print_if_term func h
    | h::t ->
      	if (func <> "print_term_type") then
	  prout_string " "; 
      	print_if_term func h ;
      	if (func <> "print_term_type") then
	  prout_string func; (* ".\n" *)
      	print_if_states_list_rec t
  in
  print_if_states_list_rec l
(* useful only when goal predicates are automatically generated:
  print_if_states_list_rec 
    (List.filter
       (function
	   Witness _
	 | Secret _ -> false
	 | _ -> true)
       l) *)
;;
(*************)

(*
   \subsection{Convert the rules section in an automat}
*)
(*
  \begin{center}
  type:
   [val print_section_rulesA : role_instance list * string -> string*int*string*(string*string*string list) list]
  \end{center}
let print_section_rules role_instancesA goal_type goal_cst =
  let filter_wanted_goals =
    List.filter (function
	Unknown(Base(Const(id)),lt) ->
	  let nid = Globals.string_id#get_name id in
	  if (nid = "request") then
            (goal_type = "request")
	    && (goal_cst = term_to_string (List.hd (List.tl (List.tl lt))))
	  else if (nid = "wrequest") then
            (goal_type = "wrequest")
	    && (goal_cst = term_to_string (List.hd (List.tl (List.tl lt))))
	  else if (nid = "witness") then
            ((goal_type = "request") or (goal_type = "wrequest"))
	    && (goal_cst = term_to_string (List.hd (List.tl (List.tl lt))))
	  else (nid <> "secret") or (goal_type = "secret")
      | _ -> true)
*)
(** print the transitions of the agents *)
let print_section_rulesA (role_instances : role_instance list)
			 (s_ids : Types.term list) (a_ids :  Types.term list) 
		 	(wa_ids :  Types.term list) : (string * int * 'd list Pervasives.ref *
							((string * string) *
								((string list * string list) * (string list * string list)) *
 					    		(string * string)) list) list =
  let filter_wanted_goals =
    let check_warning cond nid c_id =
      cond
	or (not(!Globals.flag_split)
	      && (Error_handler.displayWarning 11 (nid^" goal predicate ("
						   ^(term_to_string c_id)
						   ^") without corresp. goal property; dropped");
		  false))
    in
    List.filter (function
	Unknown(Base(Const(id)),lt) ->
	  let nid = Globals.string_id#get_name id in
	  if (nid = "request") then
(*	    List.mem (List.hd (List.tl (List.tl lt))) a_ids*)
	    let a_id = List.hd (List.tl (List.tl lt)) in
	    check_warning (List.mem a_id a_ids) nid a_id
	  else if (nid = "wrequest") then
(*	    List.mem (List.hd (List.tl (List.tl lt))) wa_ids*)
	    let wa_id = List.hd (List.tl (List.tl lt)) in
	    check_warning (List.mem wa_id wa_ids) nid wa_id
	  else if (nid = "witness") then
(*	    (List.mem (List.hd (List.tl (List.tl lt))) a_ids)
	      or (List.mem (List.hd (List.tl (List.tl lt))) wa_ids)*)
	    let a_id = List.hd (List.tl (List.tl lt)) in
	    check_warning ((List.mem a_id a_ids)
			     or (List.mem a_id wa_ids))
	      nid a_id
	  else if (nid = "secret") then
(*	    List.mem (List.hd (List.tl lt)) s_ids*)
	    let s_id = List.hd (List.tl lt) in
	    check_warning (List.mem s_id s_ids) nid s_id
	  else true
      | _ -> true)
  in
  let role_list = ref [] in
(*************)
  let stepcounter = ref 0 in
  prout_string "section rules:\n\n";
  List.iter
    (fun role_instance ->
      if not (List.exists 
		(fun role -> role=role_instance#get_name) 
		!role_list) then
	(role_list:=(role_instance#get_name)::(!role_list);
	 List.iter
	   (function
	       (_,(_, basic_r)) ->
		 let if_var_list = basic_r#get_var_list in
		 prout_string ("step step_");
		 prout_int !stepcounter; incr stepcounter;
		 prout_string " (" ; 
		 print_list term_to_string "," if_var_list; 
		 prout_string ")";
		 prout_string (" :=\n");
		 print_if_states_list ".\n" basic_r#get_lhs;
		 if (basic_r#get_cond <> []) then
		   (prout_string " &\n";
		    print_if_states_list " &\n" basic_r#get_cond);

		 let if_exists_list=basic_r#get_exists_list in
		 if (if_exists_list <> []) then
		   (prout_string "\n=[exists ";
		    print_list term_to_string "," if_exists_list;
		    prout_string "]=>\n")
		 else
		   prout_string "\n =>\n";
		 print_if_states_list ".\n"
		   (if (s_ids<>[]) or (a_ids<>[]) or (wa_ids<>[]) then
		     filter_wanted_goals basic_r#get_rhs
		   else basic_r#get_rhs);
		 prout_string "\n\n")
	 role_instance#get_actions))
    role_instances;
  prout_newline();
(***************)
   List.map
    (fun role_instance ->
      role_list:=(role_instance#get_name)::(!role_list);
       ((Globals.string_id#get_name role_instance#get_name),
        (role_instance#get_session_Id), ref [],
        List.flatten
	(List.map 
	   (function
	       (_,(_, basic_r)) ->
		(multi_transition (Globals.string_id#get_name role_instance#get_name)  (role_instance#get_session_Id)
		 (print_if_states_listA "" basic_r#get_lhs)

		 (List.map (fun e -> if_term_to_string "" e) basic_r#get_cond)
		 
		 (List.map term_to_string basic_r#get_exists_list)
		 
		(print_if_states_listA ""
		   (if (s_ids<>[]) or (a_ids<>[]) or (wa_ids<>[]) then
		     filter_wanted_goals basic_r#get_rhs
		    else basic_r#get_rhs)))
	    ) role_instance#get_actions)))
    role_instances
;;
	
(**************************************************************************)

(** !trace_story  contains all past events in the present execution of the protocol. It's can be seen like an execution trace. Each string of the list is an event in Intermediate Format *)
  let trace_story = (ref [] : string list Pervasives.ref);;
  
  (** !next_story  contains all the events executed before the user of the tool use the "Previous Step" button. It's can be seen like an execution trace. Each string of the list is an event in Intermediate Format *)
  let next_story = (ref [] : string list Pervasives.ref);;

(** contains all incoming events pruposed to the user. Each string is an event in Intermediate Format. *)
  let incomming_events = (ref [] : string list Pervasives.ref);;


let last_role = (ref [] : string list Pervasives.ref);;
let last_init = (Hashtbl.create 600 : (string, string) Hashtbl.t);;
let get_last_init role = 
	let tmp = Hashtbl.find_all last_init role in
	if([]<>tmp)
	then (car tmp)
	else ""
;;
let add_last_init role init =
	last_role := role::(!last_role);
	Hashtbl.add last_init role init
;;
let supp_last_init() =
	if([]<>(!last_role))
	then (	Hashtbl.remove last_init (car (!last_role));
		last_role := (cdr (!last_role))
	)
;;
let last_test ag1 init =(init=get_last_init ag1 || ""=init || ""=get_last_init ag1 ) ;;

(** 
update the list box lst for purpose to the user all incoming events
@return unit
*)
  let insert_lst (agent1 : string) (agent2 : string) : unit =
   let index = ref 0 in
   if(agent1<>"" || agent2<>"") then(
     List.iter 
     (fun e -> let(sender,receiver,_) = actionTreatment e in
		if(sender=agent1 || sender=agent2 || receiver=agent1 || receiver=agent2)
		then (incomming_events := remove_index (!index) (!incomming_events) ; Listbox.delete lst ~first:(`Num (!index)) ~last:(`Num (!index)) )
		else index:=1+(!index) )
   (!incomming_events)
   )
   else (Listbox.delete lst ~first:(`Num 0) ~last:`End ; incomming_events := [] );

   let not_contains_match_value state = 
   	not(List.mem false  (List.map is_a_var (cdr (decompose ("x("^state^")")))) )
   in
   let all_actions = ref [] in
   let lock = ref false in
   List.iter
   (fun (name,id,statesL,actionsL) ->
   let ag1 = (get_name_in_session name id) in
   List.iter
   (fun (state,exec) ->
   List.iter
   (fun (initial,_,final) -> 
	if((findState final)=state && (last_test ag1 (findIknows initial)) && (( ( (!exec)=true || (findIknows final)="" ) && (rcv_stack_is_empty ag1 state)) || ((!exec)=false && (List.mem (findIknows final)(get_message_stack ag1 state)))) && (operations_test_treatment_right ag1 state) && (operations_test_treatment_left ag1 state)) then(
	 List.iter
	(fun (name2,id2,statesL2,actionsL2) ->
		let ag2 = (get_name_in_session name2 id2) in
		if((name<>name2 || id<>id2) && (agent1="" || agent2="" || agent1=ag1 || agent2=ag2 || agent2=ag1 || agent1=ag2)) then(
		let prefix = ((get_name_in_session name id)^" -> "^(get_name_in_session name2 id2)^" : ") in
		List.iter
		(fun (state2,exec2) ->
		if(operations_test_treatment_right ag2 state2)
		then( 
		
			List.iter
        		(fun (initial2,_,final2) -> 
				lock:=false;
				local_variable_change:=[];
				if(state2<>(findState initial2) && match_state ag2 state2 ag2 (findState initial2))
				then(ignore(assign_var_of_state ag2 state2 (get_name_in_session name2 id2) (findState initial2)) ; clear_message_story () ; lock:=true; );
		
			(* prerr_string("\n\n"^(get_name_in_session name id)^" "^(get_name_in_session name2 id2)^"\n"^(findIknows initial2) ^"\n"^(findIknows final)^"\n"^string_of_bool((match_message (get_name_in_session name id)(findState final)(findIknows final) (get_name_in_session name2 id2)(findState initial2)(findIknows initial2)))); *)
        			if(
				(( ( (!exec2)=false  || (not_contains_match_value (findState initial2)) )
				&& (rcv_stack_is_empty ag2 state2)) || (List.mem (findIknows initial2)(get_rcv_stack ag2 state2)))
				&& ((!lock) || state2=(findState initial2) || (match_state ag2 state2 ag2 (findState initial2))
				|| state2="" )
				&& (match_message ag1 (findState final)(findIknows final) ag2 (findState initial2)(findIknows initial2)) )
				then 
				(
				ignore(assign_var ag1 (findState final)(findIknows final) ag2 (findState initial2)(findIknows initial2));
				lock:=true;

				if (operations_test_treatment_left ag2 (findState final2))
				then(
				let ic = prefix^(transcript_sender (findIknows final)) in
					if(not(List.mem ic (!all_actions)))
					then( 	all_actions:=(ic::(!all_actions));
						incomming_events := (prefix^"x("^(findIknows final)^","^((findIknows initial2))^")")::(!incomming_events);
						Listbox.insert lst ~index:(`Num 0)~texts:[(replace "_new" "" ic)]);)
				);
				if(!lock) then	(previous_message_story () ; remove_local_variable_change() );
			)actionsL2;
		
		 ) 
		)(!statesL2)
		)
   	)(!automate))
   )actionsL
   )(!statesL);
 )(!automate);
 
(* (List.iter
	(fun (name,id,statesL,actionsL) -> 
	prerr_string ((get_name_in_session name id)^"\n");
	(List.iter (fun (state2,exec2) -> prerr_string (state2^" - "^(string_of_bool (!exec2))^"\n")) !statesL);
		(List.iter 
	(fun (initial,((tg,ag),(td,ad)),final) -> prerr_string ((findState initial)^" - "^(findIknows initial)^" [ "^(list2string "," tg)^" - "^(list2string "," ag)^" - "^(list2string "," td)^" - "^(list2string "," ad)^" ] =|>\n"^(findState final)^" - "^(findIknows final)^"\n\n"))  
	actionsL);
	) (!automate));
	prerr_string "\n"; 
 *)
 (** schedule the elements in the list box lst and in the list !incomming_events *)
 index:=0;
 let tmp = ref [] in
 List.iter (fun e -> tmp:=(e,(Listbox.get lst ~index:(`Num (!index))))::(!tmp); index:=1+(!index)) (!incomming_events);
 Listbox.delete lst ~first:(`Num 0) ~last:`End;
 tmp := List.sort compare (!tmp);
 tmp := List.rev !tmp;
 incomming_events := [];
 List.iter (fun (e,l) -> incomming_events:=e::(!incomming_events) ; Listbox.insert lst ~index:(`Num 0) ~texts:[l]) (!tmp);
 
 
(*	if( (!all_actions) = [])
	then (
		let match_and_assign_with_one_state name_in_session state actionsL =
			let list2state l = List.fold_left (fun a b -> a^","^b) (car l) (cdr l) in
			let rec without_match ls1 ls2 res1 res2 = 
				match ls1 with [] -> (List.rev res1, List.rev res2)
					|(x1::l1) -> match ls2 with [] -> (List.rev res1, List.rev res2)
								|(x2::l2) -> 	if(x1=x2) then without_match l1 l2 (x1::res1) (x2::res2)
										else if(is_a_var x2) then  without_match l1 l2 (x1::res1) (x2::res2)
										else if((is_a_value x1)&&(is_a_value x2)) then  without_match l1 l2 res1 res2
										else ([""],[""])
			in
			let rec aux l =
				match l with [] -> ()
					|(((x,_),_,_)::ls) -> 
						let (s1,s2) = without_match (cdr (decompose ("x("^state^")" )))  (cdr (decompose ("x("^x^")" ))) [] [] in
						let r1 = list2state s1 in
						let r2 = list2state s2 in
						if(match_state name_in_session r1 name_in_session r2 ) 
						
				(*clear_message_story () ;*)then ignore(assign_var_of_state name_in_session  r1 name_in_session r2)
						else aux ls;
			in aux actionsL
		in
	
		List.iter 
		(fun (name,id,statesL,actionsL) -> 
			let name_in_session = get_name_in_session name id in
			List.iter (fun (state,_) -> ignore(match_and_assign_with_one_state name_in_session state actionsL) ) !statesL
		)
		(!automate)
	
	);*)
 
local_variable_change:=[];; 

let match_message_with_one_state_of_agent (sender : string) (sender_state : string Pervasives.ref) (message_sender : string) (reciever : string) (reciever_initial_state : string) : bool =
	let rec find state lActions =
		match lActions with [] -> false
			|(s::ls) -> 	if(state = (findState s) && message_sender = (findIknows s))
					then true
					else find state ls
	in
	let rec aux2 lStates lActions res =
		match lStates with [] -> List.rev res
			|((state,boolean)::ls) -> 	if(find state lActions)
							then aux2 ls lActions ((state,boolean)::res)
							else aux2 ls lActions res
	
	in
	let rec aux l_automate = 
		match l_automate with 
		((name,id,current_statesL,lActions):: ls) -> if ((get_name_in_session name id)=sender) 
						     then aux2 (!current_statesL) (List.map (fun e -> match e with (_,_,final) -> final) lActions) []
						     else aux ls;
		| [] -> failwith "Bad automate";
	in
	let rec find_states states =
		match states with  [] -> false
			|((state,exec)::ls) -> 	if( (operations_test_treatment_right sender state))
						then( 	if(!exec) then 	(
									if not(in_message_stack sender state message_sender) then
									refresh_var_of_state sender state (!tag_number); 				
									operations_affectation_treatment sender state
									);
							sender_state:=state ; 
							true
						)
						else (find_states ls)
	in
	find_states (aux (!automate));;

(** 
Refresh the automate variable after the execution of transition
@param sender is the sender of the message
@param reciever is the reciever of the message
@param message is the message
@return unit
*)
 let state_refresh (sender : string) (message : string) (reciever : string) (message_receiver : string) : unit =
   local_variable_change:=[];
   let lock_mess = ref true in
   List.iter
   (fun (name,id,statesL,actionsL) ->
   	let sender_state = ref "" in 
    	if(reciever=(get_name_in_session name id))
	then(   
		let new_statesL = ref [] in
		List.iter
		(fun (state,exec) ->
		List.iter
         	(fun (initial,_,final) ->  
			sender_state := "";  			
			if( (!exec)=false && 
				message_receiver=(findIknows initial) && 
				((match_state reciever state reciever (findState initial) ) || state=""))
			then(
				if(state<>"" && state<>(findState initial)) 
				then ( 	ignore(assign_var_of_state reciever state reciever (findState initial));
				 	clear_message_story ()
				 	) ;
				if  (match_message_with_one_state_of_agent sender sender_state message reciever (findState initial))
				then(
				(*prerr_string ("= "^reciever^" "^message^"\n"^state^" "^"\n"^(findState initial)^"\n"^(findState final)^"\n");*)
				if((findIknows initial)<>"") then (
							ignore(assign_var sender (!sender_state) message reciever (findState initial) (findIknows initial));
							);
				if(!lock_mess) then (add_last_init reciever (findIknows initial); lock_mess:=false);
				if(!intruder_mode && "(Intruder_, 0)"=reciever) then operations_affectation_treatment reciever state;
				clear_message_story ();
				local_variable_change:=[];
				let old_exec = (!exec) in
				if((findIknows final)<>"" ) && not(List.mem (findState final,ref false) (!new_statesL)) 
				then (if( (not(List.mem (findState final,ref true) (!new_statesL)))
				  (* &&  (not(List.mem (findState final,ref false) (!new_statesL))) *) )
				 then new_statesL:= (findState final,ref true)::(!new_statesL) )
				else (if( (not(List.mem (findState final,ref false) (!new_statesL))) 
				 (*&& (not(List.mem (findState final,ref true) (!new_statesL))) *) )
				then new_statesL:= (findState final,ref false)::(!new_statesL));		
				if(List.mem message (get_rcv_stack reciever state) && (findIknows initial)<>"") 
				then  (if(not(remove_rcv_stack reciever state message) && (!new_statesL)=[])
					then statesL:=(remove (state,exec) (!statesL))					
					)
				else if (not(old_exec) && (rcv_stack_is_empty reciever state)) then(
					List.iter 	
					(fun (initial0,_,final0) -> if(state=(findState final0) && message<>(findIknows initial0) && ""<>(findIknows initial0) && (findIknows initial)<>(findIknows initial0) && not(match_message reciever state message reciever (findState initial0) (findIknows initial0)) && (test_multi_rcv reciever state (findIknows initial) (findIknows initial0))  )
				 				then 	add_in_rcv_stack reciever state (findIknows initial0)
						) actionsL
					);
				)
				else if(state<>"" && state<>(findState initial)) then (previous_message_story () ; remove_local_variable_change());
			(*	List.iter
				(fun (s,e) -> 
					if(not(List.mem (s,ref true) (!new_statesL)) && not(List.mem (s,ref false) (!new_statesL)) && !e=false) then new_statesL:=((s,e)::(!new_statesL));
				)
				(!statesL);*)
			);
		 ) actionsL;
		) (!statesL);
		  remove_all_message_stack reciever;
		if ((!new_statesL)<>[]) then statesL := (!new_statesL);
		statesL := !new_statesL
	)
     )(!automate);
	
     	List.iter
	(fun (name,id,statesL,actionsL) ->
		if(sender=(get_name_in_session name id))
		then(
			let lock = ref true in
			let new_statesL = ref [] in
			List.iter
			(fun (state,exec) -> 
			if(operations_test_treatment_right sender state)
			then(
				if ( (!exec) && (message_stack_is_empty sender state))
		 		then (  
					List.iter
     	   				(fun (initial,_,final) -> if(state=(findState final) &&  ""<>(findIknows final) )
								then (	exec:=false;
									if((!lock) && ""=(findIknows initial) && message=(findIknows final)) 
									then (lock:=false; (*new_statesL:=[] ;*)  statesL:=[] ; remove_all_message_stack sender );
									);
					)
					actionsL;
				
					List.iter
        				(fun (initial,_,final) -> if( ((!lock) || (findIknows initial)="") && state=(findState final))
								then (	if(message<>(findIknows final)) 
									then add_in_message_stack sender state (findIknows final);
									if((!lock) && (findIknows initial)="")
									then 	(if(not(List.mem (findState final,ref true) (!new_statesL)))
										then new_statesL:= (findState final,ref true)::(!new_statesL))
									else 	(if(not(List.mem (findState final,ref false) (!new_statesL)))
										then new_statesL:= (findState final,ref false)::(!new_statesL));
								)
					)
					actionsL;			
					)
				else if(not(!exec) && (*(!new_statesL)=[] &&*) (message_stack_is_empty sender state)) 
					then	statesL:=(remove (state,exec) (!statesL))
				else if(not(message_stack_is_empty sender state)) 
					then  (if(not(remove_message_stack sender state message) (* &&  (!new_statesL)=[] && (!exec) *))
						then statesL:=(remove (state,exec) (!statesL))					
					);			
				 ) 
			)
			(!statesL);

			remove_all_rcv_stack sender;

			if ((!new_statesL)<>[]) then statesL := (remove_double ((!new_statesL)@(!statesL)));
			(*List.iter
			(fun (s,e) ->
				if(!e=true && (List.mem (s, ref false) (!statesL)) ) then statesL := remove (s,e) (!statesL)
			)
			(!new_statesL) *)
			);
			
	        if((*sender=(get_name_in_session name id) || *)reciever=(get_name_in_session name id))
		then(
			let succ_state = ref ([] : (string * bool ref ) list) in
			List.iter
			(fun (state,exec) -> 
			List.iter
        		(fun (initial,_,final) -> 
				succ_state:= (if (((!exec)=false && (List.mem (findIknows final)(get_message_stack reciever state))))
				then []
				else (let rec aux l res=
					match l with [] -> List.rev res;
						|((initial1,_,final1)::ls) ->
							if ((match_state reciever state reciever (findState initial1))
								&& (findIknows initial1)=""							
								&& not(List.mem ((findState final1),ref false) (!statesL)) 
								&& not(List.mem ((findState final1),ref false) (!succ_state)) 
								&& not(List.mem ((findState final1),ref false) res))
							then 	(aux ls (((findState final1),ref true)::res))
							else aux ls res ;
					in
					aux actionsL [] ))@(!succ_state);
			)actionsL;
			) (!statesL);
			if ((!succ_state)<>[]) then statesL := (remove_double ((!statesL)@(!succ_state)));
		);
			
				
		) (!automate);
   
	(*Hashtbl.iter 
	(fun  (name_in_session,state_t,type_of_message) m -> 
		if((type_of_message="_message_stack") && (name_in_session=sender) ) 
		then refresh_var_of_state name_in_session state_t (!tag_number))
	message_stack; *)
   	local_variable_change:=[];
	if(!intruder_mode && ("(Intruder_, 0)"=reciever || "(Intruder_, 0)"=sender)) then update_intruder_knowledge ()
   ;;   
   
   
   	
(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
(** Widgets for intruder mode *)

(** frame for the "window_compose_intruder_knowledge" button and the "label_intruder" label in the intruder knowledge *)
let intruderFrameForButton = (Frame.create ~width:20 ~borderwidth:0 infoFrame : Widget.frame Widget.widget);;
(** frame for display intuder knowledge and for the button for compose a message with the intruder knowledge *)
let intruderFrame = (Frame.create ~width:20  infoFrame : Widget.frame Widget.widget);;
(** a listbox for display the intruder knowledge *)
let intruder_knowledge_lst = (Listbox.create ~width:15 ~height:10 ~background:clicol  ~selectmode:`Single intruderFrame : Widget.listbox Widget.widget);;
(** scroll bar vertical of intruder_knowledge_lst *)
let intruder_knowledge_v = (Scrollbar.create ~background:violet ~orient:`Vertical  intruderFrame :  Widget.scrollbar Widget.widget);;
(** scroll bar horizontal of intruder_knowledge_lst *)
let intruder_knowledge_h = (Scrollbar.create ~background:violet ~orient:`Horizontal  intruderFrame :  Widget.scrollbar Widget.widget);;
  scroll_link intruder_knowledge_v intruder_knowledge_lst "vertical";;
  scroll_link intruder_knowledge_h intruder_knowledge_lst "horizontal";;
(**
 display the intruder knowledge in the list box intruder_knowledge_lst
 @return unit
*)
 let display_intruder_knowledge (() : unit) : unit =
	Listbox.delete intruder_knowledge_lst ~first:(`Num 0) ~last:`End;
	let knowledge = let tmp = decompose(get_value_of_var "x" "set_knowledge") in
			if(tmp<>[]) then (cdr tmp)
			else []
	in
	Listbox.insert intruder_knowledge_lst ~index:`End ~texts:(List.map transcript_sender knowledge)
 ;;
  (**
 create a label in the intruder knowledge
 *)
let label_intruder = Label.create ~text:"Intruder knowledge : "  ~borderwidth:1 ~relief:`Ridge intruderFrameForButton;;
 (**
 create a window for compose a message with the intruder knowledge
 *)
let window_compose_intruder_knowledge = (Button.create ~text:"Compose knowledge" ~borderwidth:1 ~height:1 ~relief:`Raised intruderFrameForButton 
	~command:(fun () -> window_compose_intruder_knowledge root trace_story incomming_events (insert_lst : string -> string -> unit) (display_intruder_knowledge : unit -> unit)) : Widget.button Widget.widget);;
(* ***** ***************** ***** *)
 
(** click is used for create a double-click.*)
 let click = (ref (-1) : int Pervasives.ref);;
   
(** Compute the string for display the message send on the network during the execution of a step 
@param lst listbox lst
@param sender is the name of the sender of the message
@param message is the message send by the sender and receive by another agent
@return string sdqf
*)
 let emission (lst :  Widget.listbox Widget.widget) (sender : string) (message : string) : string =
 	let res =ref ("["^message^"]") in
	let l = ref [message] in
	let message_if = transcript_sender message in
	Hashtbl.iter
	(fun (name_in_session,_,type_of_message) b_arg ->
		let b = transcript_sender b_arg in
		if(	b<>""
			&&(not (List.mem b (!l)))
			&& (sender=name_in_session)
			&& (type_of_message="_message_stack" )
			&& (let rec test_no_recieve_message l_automate =
				match l_automate with [] -> false
						|((name,id,_,actionsL)::ls) -> 
							if(sender=(get_name_in_session name id)) 
							then ( 
								let rec aux l =
									match l with [] -> false
										|((initial,_,final)::ll) -> 
											if(	message_if=(findIknows final) &&
												""<>(findIknows initial) )
											then true
											else aux ll;
							in aux actionsL
						)
						else test_no_recieve_message ls
				in test_no_recieve_message (!automate)) 
(*			&& not(List.mem true 
				(List.map
					(fun action_lst ->
						let (sender_lst, _, message_lst) = actionTreatment action_lst in
						(sender_lst=sender && message_lst=b))
						(Listbox.get_range lst ~first:(`Num 0) ~last:`End)
			)) *)
		)
		then (l:=(b::(!l)) ; res:=(!res)^"["^b^"]" ))
	message_stack;
	if((!res)<>"") 
	then "...Emission"^(replace "_new" "" (!res))^"!!!..."
	else ""
 ;;

 (** Text variable for the check button "Untyped variables" *)
 let tv_untypedButton = (Textvariable.create () : Textvariable.textVariable);;
 Textvariable.set tv_untypedButton "0";;
 (** checkbutton "Untyped variables" *)
 let untypedButton = (Checkbutton.create ~text:"Untype" 
 	~borderwidth:1 
	~onvalue:"1" ~offvalue:"0" ~variable:tv_untypedButton ~height:1
	~command:(fun() -> 
			click := (-1); 
			if(!untyped_variable) 
			then untyped_variable:=false 
			else untyped_variable:=true) buttonFrame :  Widget.checkbutton Widget.widget);; 
(** Button "Next Step" *)	
 let goButton = (Button.create ~text:"Next step >" ~width:10
 		~state:`Disabled
	~command:(fun () -> 
			let action = if((!next_story)<>[]) then car !next_story else "" in
			let other_action = if(!intruder_mode && (not(containStr "->" action))) then decompose action else [] in
			if(other_action<>[] && ((!next_story)<>[])) then(
				if((car other_action)="add_in_set") then add_in_set (cadr other_action) (caddr other_action) (caddr (cdr other_action)) (caddr (cddr other_action));
				next_story := (cdr !next_story);
				trace_story := action::(!trace_story);
				if(!intruder_mode && containStr "(Intruder_, 0)" action) 
				then (update_intruder_knowledge() ; display_intruder_knowledge();)
			)
			else if ((!next_story)<>[]) then(
				next_story := (cdr !next_story);
				trace_story := action::(!trace_story);
				let (sender, reciever, message_arg) = actionTreatment action in
				let message = decompose message_arg in
				add_in_automate_story ();
				add_in_variable_story ();
				next_message_stack ();
				next_id_refresh_story ();
				state_refresh sender (cadr message) reciever (caddr message);
				add_tag_number ();
				clear_message_story ();
				if(!intruder_mode) then display_intruder_knowledge();
				let message_sender = replace "_new" "" (transcript_sender (cadr message)) in
				let message_receiver = replace "_new" "" (transcript_receiver (caddr message)) in
				let message_real = transcript_to_real sender (cadr message) in
				Listbox.insert story ~index:`End ~texts:[sender^" -> "^reciever^" : "^message_sender];
				draw_action sender reciever red message_sender message_real message_receiver black (emission lst sender message_sender) green;
				insert_lst sender reciever;
				)
) buttonFrame : Widget.button Widget.widget);;

(** Button "Previous Step" *)
let previousButton = (Button.create ~text:"< Previous step" ~width:10
	~command:(fun() -> 
	let rec trace_story_adjust () =
			match (!trace_story) with [] -> ()
				|(x::ls) -> 	if(not(containStr "->" x))
						then (next_story:=(car (!trace_story))::(!next_story) ; trace_story := ls ; trace_story_adjust ());
		in 

	if((!tag_number)>0) then(
		supp_last_init ();
		Button.configure goButton ~state:`Normal;
		remove_graphics_obj ();
		previous_id_refresh_story ();
		previous_tag_number ();
		previous_yAction ();
		clear_message_story ();
		previous_state();
		previous_variable_state ();
		previous_message_stack ();
		if(!intruder_mode) then (trace_story_adjust (); update_intruder_knowledge () ; display_intruder_knowledge());
		let (sender, receiver, _) = actionTreatment  (car (!trace_story)) in
		next_story:=(car (!trace_story))::(!next_story);
		trace_story := (cdr (!trace_story));
		insert_lst sender receiver ;
		Listbox.delete story ~first:`End ~last:`End;
		click := (-1))
	else 	if(!intruder_mode && (!tag_number)=0 && (!trace_story)<>[]) then(
		Button.configure goButton ~state:`Normal;
		next_story:=(car (!trace_story))::(!next_story);
		trace_story := (cdr (!trace_story));
		previous_state();
		previous_message_stack ();
		previous_variable_state ();
		trace_story_adjust (); 
		update_intruder_knowledge () ; 
		display_intruder_knowledge();
		);
	)
	buttonFrame : Widget.button Widget.widget);;

pack ~side:`Left ~fill:`Both  ~expand:true [previousButton];;
pack ~side:`Left ~fill:`Both  ~expand:true [goButton];;
pack ~side:`Left ~fill:`Both  ~expand:true [untypedButton];;

  bind lst
		  ~events:[(`ButtonPressDetail 1)]
		  ~fields:[`MouseY]
		  ~action:(function x -> 
			     let result = 
			       Listbox.nearest 
				 lst
			       ~y:x.ev_MouseY
			     in
			     let select = (* obtain the integer from the [`Num of int] *)
			       match result with
				 | `Num num -> num
			     in
			        if((!click)=select)
				then(
				let action = if ((List.length (!incomming_events))>0) then List.nth (!incomming_events) select else "" in
				if(action<>"") then (
					Button.configure goButton ~state:`Disabled;
					next_story:=[];
					Listbox.insert story ~index:`End ~texts:[Listbox.get lst ~index:(`Num select)];
					trace_story := action::(!trace_story);
					let (sender, reciever, message) = actionTreatment action in
					let m = decompose message in
					add_in_automate_story ();
					add_in_variable_story ();
					next_message_stack ();
					(*view_table_var ();*)
					next_id_refresh_story ();
					state_refresh sender (cadr m) reciever (caddr m);
					 if(!intruder_mode) then display_intruder_knowledge();
					add_tag_number ();
					(*view_table_var ();*)
					clear_message_story ();
					let message_sender = transcript_sender (replace "_new" "" (cadr m)) in
					let message_receiver =  transcript_receiver (replace "_new" "" (caddr m)) in
					let message_real = transcript_to_real sender (cadr m) in
					draw_action sender reciever red message_sender message_real message_receiver black (emission lst sender message_sender) green;
					insert_lst sender reciever;
					(*view_table_var ()*)
					);
				click := (-1)
				)
				else click := select;
			);;
 (** Variables manitoring *)

(** role_click_zone draw the zone for which the user can click and see value of variables of the role
@param name of the role
@return unit
*)
let draw_role_click_zone (name : string) : unit =
	let role_x = Hashtbl.find role_position name
	in
	let rec aux y_pos tag = 
		if(tag>0)
		then (  
			if(in_zone y_pos)
			then aux (y_pos+. !yDelta ) tag
			else (
				draw_click_zone role_x y_pos tag (name^"click_zone");
				aux (y_pos+. !yDelta ) (tag - 1)
			);
		);
	in
	let rec aux2 y_pos tag = 
		if(tag>0)
		then (  
			if(in_zone y_pos)
			then aux2 (y_pos+. !yDelta ) tag
			else (
				if (in_click_zone role_x y_pos) then draw_click_zone role_x y_pos tag (name^"click_zone");
				aux2 (y_pos+. !yDelta ) (tag - 1)
			);
		);
	in
	remove_graphics_obj_with_tag_value (name^"click_zone");
	if( (Hashtbl.find_all var_in_monitor name)<>[] )  then aux (!yOrd +. !yDelta) !tag_number
	else aux2 (!yOrd +. !yDelta) !tag_number
;;

(** list of pair : textvariable and checkbutton for create a frame for select the name of the variables for which the user wants toknow the value.
Each element is a checkbutton of the frame and his Textvariable *)			
let button_list = (ref [] : (Textvariable.textVariable * Widget.checkbutton Widget.widget) list Pervasives.ref);;
(** var_frame create a frale with button for select the variable of a role for which the user wants to see the value
@param name of the role
@return unit
*)
let var_frame (name : string) : unit  =
	button_list := [];
	let root_var =  Toplevel.create ~borderwidth:1 ~relief:`Groove root in
	Wm.title_set root_var ("Variables of "^name);
	let frame_var = Frame.create ~borderwidth:1 ~relief:`Sunken root_var in
	let frame_button = Frame.create ~borderwidth:1 ~relief:`Ridge root_var in
	let label_blank_top= Label.create ~width:32 ~height:1 root_var in
	let label_blank_middle= Label.create ~width:32 ~height:1 root_var in
	let label_blank_bottom= Label.create ~width:32 ~height:1 root_var in
	List.iter
	(fun a -> 
		let tv = Textvariable.create () in
		if (List.mem a (Hashtbl.find_all var_in_monitor name)) then Textvariable.set tv "1" else Textvariable.set tv "0";
		button_list := (tv,(
			Checkbutton.create ~text:a ~width:26 ~height:2 ~borderwidth:1 ~indicatoron:false ~selectcolor:clicol
			~onvalue:"1" ~offvalue:"0" ~variable:tv			 
			~command:(fun () -> 
				let l_var = Hashtbl.find_all var_in_monitor name in
				if (List.mem a l_var) 
				then (
				List.iter (fun _ -> Hashtbl.remove var_in_monitor name) l_var;
				List.iter 
				(fun t  -> (Hashtbl.add var_in_monitor name t))
				(List.rev (remove a l_var))
			)
			else (Hashtbl.add var_in_monitor name a)		
		)  frame_var))::(!button_list)
	)
	(Hashtbl.find_all role_var_monitoring name);
	let button_unselect_all = Button.create  ~text:"Unelect all" 
	~command:(fun () ->  	List.iter
				(fun (tv,w) -> if(Textvariable.get tv="1") then ((*Textvariable.set tv "0" ; *)Checkbutton.invoke w))
				(!button_list)
		)
	~width:25 ~height:1 ~borderwidth:1 ~relief:`Groove frame_button 
	in 
	let button_select_all = Button.create  ~text:"Select all" 
		~command:(fun () ->  	List.iter
					(fun (tv,w) -> if(Textvariable.get tv="0") then ((*Textvariable.set tv "1" ;*) Checkbutton.invoke w))
					(!button_list)
		)
		~width:25 ~height:1 ~borderwidth:1 ~relief:`Groove frame_button 
	in 
	let button_ok = Button.create  ~text:"OK" 
		~command:(fun () -> ignore(
				List.iter (fun (_,obj) -> destroy obj)(!button_list);
				destroy frame_var;
				destroy button_unselect_all;
				destroy button_select_all;
				destroy frame_button;
				destroy root_var;
				draw_role_click_zone name
			)
		)
		~width:25 ~height:1 ~borderwidth:1 ~relief:`Groove frame_button 
	in 
	pack ~side:`Top ~fill:`Y [label_blank_top];
	pack ~side:`Top ~fill:`Y [frame_var];
	pack ~side:`Top ~fill:`Y [label_blank_middle];
	pack ~side:`Top ~fill:`Y [frame_button];
	pack ~side:`Top ~fill:`Y [label_blank_bottom];
	List.iter
	(fun (_,check_button) -> pack ~side:`Top ~fill:`Y [check_button])
	(!button_list);
	pack ~side:`Top ~fill:`Y [button_unselect_all];
	pack ~side:`Top ~fill:`Y [button_select_all];
	pack ~side:`Top ~fill:`Y [button_ok]
;;

bind    
  ~events:[(`ButtonPressDetail 1)]
  ~fields:[`MouseX;`MouseY]
  ~extend: true
  ~action: 
    (function e -> 
    	let round100 x = (round (x *. 100.)) /. 100. in
    	let originex (a,_) =   (round100 a) *. float_of_int (
	((xconv (float_of_int(!nbAgents + 1) *. 10. +. 5. )) + (5*(get_max_length_message())))
	- 
	((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) ) ))
	+.
	( (!xscale) *. float_of_int((-1) * 10) -. float_of_int (5*(get_max_length_message()) ))
	in
	let originey (a,_) =   (round100 a) *. (float_of_int (max (yconv !yOrd) (yconv !ymin))) in
    	let x_scroll1 = originex (Canvas.xview_get canvas) in
    	let y_scroll1 = originey (Canvas.yview_get canvas) in
    	let x = ((((float_of_int e.ev_MouseX) +. x_scroll1)/.(!xscale)) +. (!xmin)) in
	let y = round100( (!ymax) -. (((float_of_int e.ev_MouseY) +. y_scroll1 )/.(!yscale)) ) in
(*	 prerr_string((string_of_int(e.ev_MouseX))^" - "^(string_of_int(e.ev_MouseY))^"\n");
	 	 prerr_string((string_of_float x)^" -- "^(string_of_float y)^"\n");
	prerr_string((string_of_float(x_scroll1))^" = "^(string_of_float(y_scroll1))^" = "^"\n"); *)
(*	Canvas.create_oval ~x1:(xconv x) ~y1:(yconv (!yOrd)) ~x2:(xconv x) ~y2:(yconv y)  canvas;
	Canvas.create_line ~xys:[((xconv (x +. 0.1)),(yconv (y+.0.1)));((xconv x),(yconv y))] canvas;*)
	let role_x = ref (-1000.) in
	let role_name = ref "" in
	let tag_pos = ref (-1000.) in
	Hashtbl.iter (fun r pos ->
		if( ((pos +. 1.) > x) && ((pos -. 1.) < x) ) then (role_name := r ; role_x:=pos)
	)role_position;
	
	let n = !ymax -. !yMarge -. y in
	let n_round = (round (n /. !yDelta)) *. !yDelta in
	if ((abs_float (n -. n_round))< ((!yDelta)/. 3.)) then tag_pos:=  !ymax -. !yMarge -. n_round;
(* ------------------------ *)
	let n_var = ref (List.length (Hashtbl.find_all var_in_monitor (!role_name))) in
	let n_l_intersec = (number_of_common_line_between_all_zones (!role_x) (!tag_pos) ) in
	let n_line = ref (!n_var - (min !n_var n_l_intersec)) in
	let old_yOrd = !yOrd in
	
	let rec aux y_pos tag = 
		let rec return_to_new_line n_line = 
			if(n_line>0) then (
				Canvas.move canvas (`Tag ("tag"^(string_of_int tag))) 0  (int_of_float ((!yscale)*.(!yDelta)) ) ; 
				return_to_new_line (n_line - 1)
			)
		in
		if(y_pos  > old_yOrd && tag>0)
		then ( 
			if(in_zone y_pos)
			then aux (y_pos-. !yDelta ) tag
			else (	return_to_new_line !n_line;
 				aux (y_pos-. !yDelta ) (tag - 1)
			) 
		)
		else (
			if not(in_zone y_pos)
			then return_to_new_line !n_line;
			tag;
		);
	in
	let rec get_tag y_pos tag = 
		if(y_pos  > old_yOrd && tag>0)
		then (  
			if(in_zone y_pos)
			then get_tag (y_pos-. !yDelta ) tag
			else get_tag (y_pos-. !yDelta ) (tag - 1);
		)
		else tag;
	in
	let lock = (!role_name)<>""  && not(in_zone !tag_pos) && !tag_pos > old_yOrd && !tag_pos <= (!ymax -. !yMarge -. !yDelta) in
	let (tag_insert, in_zone_bool)= 	
				if(((!role_name)<>"") && not(in_click_zone (!role_x) (!tag_pos)) && lock ) 
				then ((aux (!tag_pos -. !yDelta) (!tag_number)), true)
				else ((get_tag (!tag_pos -. !yDelta) (!tag_number)), false);
	in
	

	if(in_zone_bool && lock && !n_var > 0)
	then (
		add_in_zone (!role_x) (!tag_pos)(* -. ((float_of_int !n_line ) *. !yDelta))*) [(((!tag_pos) -. ((float_of_int (!n_line - !n_var)) *. !yDelta) +. (!yDelta *. 0.5)))  ; !tag_pos -. ((float_of_int !n_line ) *. !yDelta) +. (!yDelta *. 0.5) ]
		)
	else if( (!role_name)<>""  && !tag_pos > old_yOrd && !tag_pos <= (!ymax -. !yMarge -. !yDelta) )
	     then (
		n_var := find_number_of_var_in_monitor !role_x !tag_pos ; 
		n_line := (!n_var - (min !n_var n_l_intersec)); 
		if(!n_var > 0) then del_zone (!role_x) (!tag_pos)
		);

	  let rec draw_variables list_varName y_pos =
	    (* debug let _ = view_table_var () in *)
		if (list_varName<>[])
		then(	
			let varName = (car list_varName) in
			let var = replace (!role_name) "" varName in 

(* Variable monitoring display  *)
(* get_value_of_var_step 3rd field: protocol step *)
			draw_text_left (!role_x +. 0.0) y_pos (var^" : "^(transcript_sender (get_value_of_var_at_step (!role_name) var (tag_insert+2)))) violet  tag_insert ((!role_name)^"variablesMonitoring") ;
			draw_variables (cdr list_varName) (y_pos +. !yDelta)
		)
	in 
	
	if (not(in_zone_bool) && lock && !n_var > 0)
	then 	( 	remove_graphics_obj_with_tag  tag_insert  ((!role_name)^"variablesMonitoring"); 
			yOrd:= !yOrd +. (float_of_int !n_line) *. !yDelta; 
			let rec aux2 tag = 
				if(tag<=(!tag_number))
				then (
				let rec return_to_new_line n_line = 
					if(n_line>0) then (
						Canvas.move canvas (`Tag ("tag"^(string_of_int tag))) 0  (int_of_float ((-1.)*.(!yscale)*.(!yDelta)) ) ; 
						return_to_new_line (n_line - 1)
					)
				in return_to_new_line !n_line;
				aux2 (tag + 1)
				)
			in if(!n_line >0) then (aux2 (tag_insert+0));
			Canvas.configure canvas ~scrollregion:((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) ),0,(xconv ((float_of_int(!nbAgents + 1)) *. 10. +. 5. )) + (5*(get_max_length_message())), (yconv !yOrd));
			draw_role_click_zone (!role_name)
		)
	else if ( lock && (Hashtbl.find_all var_in_monitor (!role_name))<>[] ) 
	then (	draw_variables (Hashtbl.find_all var_in_monitor (!role_name)) ((!tag_pos -. ((float_of_int !n_line) *. !yDelta)) +. !yDelta); yOrd:= !yOrd -. (float_of_int !n_line) *. !yDelta;
		Canvas.configure canvas ~scrollregion:((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) ),0,(xconv ((float_of_int(!nbAgents + 1)) *. 10. +. 5. )) + (5*(get_max_length_message())), (yconv !yOrd))
	) ; 
	
  )
  canvas
;; 
			

(* %%%%%%%%%%%%%%%%%%%%%%%%%% *)

(** file of the HLPSL spceicification of the protocol *)
let fileName = (ref "" : string Pervasives.ref);;
(** file of the execution trace of the protocol *)
let traceFileName = (ref "" : string Pervasives.ref);;
 
(** Menu and Menubutton *)

(** Menu_button "Trace Files" *)
let meb = (Menubutton.create ~text:"Trace Files" ~font:"guifont" ~activebackground:blue_light bar : Widget.menubutton Widget.widget);;
(** Menu "Trace Files" *)
let men = (Menu.create meb ~activebackground:blue_light ~borderwidth:1 ~relief:`Groove : Widget.menu Widget.widget);;
Menu.add_command ~label:"Open trace..." ~command:(fun () -> 
	let file = (!traceFileName) in
	traceFileName:=(Tk.getOpenFile() ~defaultextension:".trace" ~filetypes:types ~initialdir:(Filename.dirname (!fileName)) ~title:"Open a trace file");
	if((!traceFileName)<>"") 
	then (
		Wm.title_set root (sim_version^" : "^(Filename.basename (!fileName))^" - Trace File : "^ (Filename.basename (!traceFileName)));
		let list_file = readFile (!traceFileName) in
		Listbox.delete lst ~first:(`Num 0) ~last:`End;
		Listbox.delete story ~first:(`Num 0) ~last:`End;

		let rec trace_story_adjust () =
			match (!trace_story) with [] -> ()
				|(x::ls) -> 	if(not(containStr "->" x))
						then (trace_story := ls ; trace_story_adjust ());
		in 
		let rec aux () = 
			if (!tag_number > 0) 
			then (
				remove_graphics_obj ();
				previous_id_refresh_story ();
				previous_tag_number ();
				previous_yAction ();
				clear_message_story ();
				previous_state();
				previous_variable_state ();
				previous_message_stack ();
				 if(!intruder_mode) then (trace_story_adjust () ; update_intruder_knowledge ()) ;
				trace_story := (cdr (!trace_story));
				aux ()
			)
			else 	if(!intruder_mode && (!tag_number)=0  && (!trace_story)<>[]) then(
					previous_state();
					previous_variable_state ();
					previous_message_stack ();
					trace_story_adjust (); 
					update_intruder_knowledge () ; 
					display_intruder_knowledge();
				);
		in aux ();
		List.iter
		(fun action ->
		if(action<>"") 
		then(	let other_action = if(!intruder_mode && (not(containStr "->" action))) then decompose action else [] in
			if(other_action<>[]) then(
				if((car other_action)="add_in_set") then add_in_set (cadr other_action) (caddr other_action) (caddr (cdr other_action)) (caddr (cddr other_action));
				trace_story := action::(!trace_story);
				if(!intruder_mode && containStr "(Intruder_, 0)" action) then update_intruder_knowledge();
			)
			else(
				trace_story := action::(!trace_story);
				let (sender, reciever, message_arg) = actionTreatment action in
				let message = decompose message_arg in
				add_in_automate_story ();
				add_in_variable_story ();
				next_message_stack ();
				next_id_refresh_story ();
				state_refresh sender (cadr message) reciever (caddr message);
				add_tag_number ();
				clear_message_story ();
				let message_sender = replace "_new" "" (transcript_sender (cadr message)) in
				let message_receiver = replace "_new" "" (transcript_receiver (caddr message)) in
				let message_real = transcript_to_real sender (cadr message) in
				Listbox.insert story ~index:`End ~texts:[sender^" -> "^reciever^" : "^message_sender];
				draw_action sender reciever red message_sender message_real message_receiver black (emission lst sender message_sender) green;
				)
			);  
		) list_file;
		Button.configure goButton ~state:`Disabled;
		insert_lst "" "";		
		if(!intruder_mode) then display_intruder_knowledge ();
	)
	else traceFileName:=file;
) men;;


Menu.add_command ~label:"Save trace" ~command:(fun () -> (
	if ((!traceFileName)<>"") 
	then(
		let list_trace = 	List.map
					(fun action -> 
						if(!intruder_mode && (not(containStr "->" action))) then action
						else	(let (sender, reciever, message_arg) = actionTreatment action in
							(sender^" -> "^reciever^" : "^message_arg)); )
					(List.rev !trace_story)
					in
		let txt_trace =  List.fold_left  (fun a b -> a^"\n"^b) "" list_trace in
		saveFile (!traceFileName) txt_trace
	)
	else(
		let file = (!traceFileName) in
		traceFileName := 
		(Tk.getSaveFile() ~defaultextension:".trace" ~filetypes:types ~initialdir:(Filename.dirname (!fileName)) ~title:"Save the trace file";);
		if((!traceFileName)<>"")
		then(
			Wm.title_set root (sim_version^" : "^(Filename.basename (!fileName))^" - Trace File : "^ (Filename.basename (!traceFileName)));
			let list_trace = List.rev !trace_story in
			let txt_trace =  List.fold_left  (fun a b -> a^"\n"^b) "" list_trace in
			saveFile (!traceFileName) txt_trace
		)
		else traceFileName:=file
	)
)) men;;
Menu.add_command ~label:"Save trace as..." ~command:(fun () ->  
	let file = (!traceFileName) in
	traceFileName := 
		(Tk.getSaveFile() ~defaultextension:".trace" ~filetypes:types ~initialdir:(Filename.dirname (!fileName)) ~title:"Save the trace file";);
	if((!traceFileName)<>"")
	then(
		Wm.title_set root (sim_version^" : "^(Filename.basename (!fileName))^" - Trace File : "^ (Filename.basename (!traceFileName)));
		let list_trace = List.rev !trace_story in
		let txt_trace =  List.fold_left  (fun a b -> a^"\n"^b) "" list_trace in
		saveFile (!traceFileName) txt_trace
	)
	else traceFileName:=file
) men;;
Menu.add_separator men;;
Menu.add_command ~label:"Export schema as..." ~command:(fun () ->
	let file = 
		(Tk.getSaveFile() ~defaultextension:".ps" ~filetypes:types_schema ~initialdir:(Filename.dirname (!fileName)) ~title:"Export schema as...";) 
	in
	if(file<>"")
	then (
	let temp_dir =(Utils_simulator.path_span^"/testsuite/results/") in
	let temp_file = (temp_dir^(Filename.basename file)) in
	if(Sys.file_exists temp_file) then ignore(Sys.remove temp_file);
	let canvas_width = ((xconv ((float_of_int(!nbAgents + 1)) *. 10. +. 5. )) + (5*(get_max_length_message()))) - ((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) )) in
	let canvas_height = (yconv !yOrd) in
	saveFile temp_file (Canvas.postscript canvas ~x:((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) )) ~y:0  ~pagewidth:(int_of_float ((250. *. (float_of_int canvas_width))/. (!width))) ~pageheight:(int_of_float ((800. *. (float_of_int canvas_height))/. (!height)))
	~width:canvas_width ~height:canvas_height );
			
	 if(Filename.check_suffix file ".ps") then ignore(Sys.command(esc^"ps2ps "^(Filename.quote temp_file)^" "^(Filename.quote file)^esc))
	 else if(Filename.check_suffix file ".pdf") then( 
		let temp_file_ps_aux = (temp_dir^(Filename.basename ((Filename.chop_suffix temp_file ".pdf")^".ps"))) in
		if(Sys.file_exists temp_file_ps_aux) then ignore(Sys.remove temp_file_ps_aux);	 
		ignore(Sys.command(esc^"ps2ps "^(Filename.quote temp_file)^" "^(Filename.quote temp_file_ps_aux)^esc));
		ignore(Sys.command(esc^"ps2pdf "^(Filename.quote temp_file_ps_aux)^" "^(Filename.quote file)^esc));
		if(Sys.file_exists temp_file_ps_aux) then ignore(Sys.remove temp_file_ps_aux);	
	 );
	if(Sys.file_exists temp_file) then ignore(Sys.remove temp_file);
	)
) men;;
Menu.add_command ~label:"Print schema..." ~command:(fun () ->

	let file_config_command = (Utils_simulator.path_span^"/bin/protocol_simulation/print_command.ini") in
	let command = ref(	if(Sys.file_exists file_config_command)
				then(	let rec aux l =
						match l with [] -> "lpr"
							|(x::ls) -> 	if(x<>"")
									then x
									else aux ls;
					in aux (readFile file_config_command)
				)
				else "lpr" );
	in
	
	let print () =
		let temp_dir =(Utils_simulator.path_span^"/testsuite/results/") in
		let temp_file = (temp_dir^"print_temp.ps") in
		if(Sys.file_exists temp_file) then ignore(Sys.remove temp_file);
		let canvas_width = ((xconv ((float_of_int(!nbAgents + 1)) *. 10. +. 5. )) + (5*(get_max_length_message()))) - ((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) )) in
		let canvas_height = (yconv !yOrd) in
		saveFile temp_file
			(Canvas.postscript canvas ~x:((xconv ((-1.) *. 10. -. 5. )) - (5*(get_max_length_message()) )) ~y:0  ~pagewidth:(int_of_float ((250. *. (float_of_int canvas_width))/. (!width))) ~pageheight:(int_of_float ((800. *. (float_of_int canvas_height))/. (!height)))
	~width:canvas_width ~height:canvas_height );
		let temp_file_ps_aux = (temp_dir^"print.ps") in
		if(Sys.file_exists temp_file_ps_aux) then ignore(Sys.remove temp_file_ps_aux);	
		ignore(Sys.command(esc^"ps2ps "^(Filename.quote temp_file)^" "^(Filename.quote temp_file_ps_aux)^esc));
		if(Sys.file_exists temp_file) then ignore(Sys.remove temp_file); 
		(*ignore(Thread.create (fun () -> ignore( *)
		ignore(Sys.command(esc^(!command)^" "^(Filename.quote temp_file_ps_aux)^" &"^esc));
							(*if(Sys.file_exists temp_file_ps_aux) then ignore(Sys.remove temp_file_ps_aux))) 
				() );
		Thread.delay 1. *)
	in
	
	let root_print_command =  Toplevel.create ~borderwidth:1 ~relief:`Groove root in
	Wm.title_set root_print_command ("Print command ...");
	let frame_print_command1 = Frame.create ~borderwidth:1 ~relief:`Sunken root_print_command in
	let frame_print_command2 = Frame.create ~borderwidth:1 ~relief:`Sunken root_print_command in
	let label_print_command = Label.create ~text:"Printer Command : " ~width:15 ~height:2 frame_print_command1 in
	let entry_print_command = Entry.create ~width:20 ~borderwidth:1 frame_print_command1 ~background:selcol in
	Entry.insert entry_print_command ~index:`End ~text:(!command) ;
	let button_ok_print_command = 	Button.create ~text:"OK" 
						~command:(fun () -> 	
									if((Entry.get entry_print_command)<>"")
									then (	if(Sys.file_exists file_config_command) 
										then ignore(Sys.remove file_config_command);
										command := (Entry.get entry_print_command);
										saveFile file_config_command (!command)
										);
										
									destroy label_print_command ; destroy entry_print_command ; 
									destroy entry_print_command ; destroy root_print_command;
									print () ) 
					~width:19 ~height:1 ~borderwidth:1 ~relief:`Groove frame_print_command2  in
	let button_cancel_print_command = Button.create ~text:"Cancel" 
						~command:(fun () -> 	destroy label_print_command ; destroy entry_print_command ; 
									destroy button_ok_print_command ; destroy entry_print_command ; 
									destroy root_print_command ) 
						~width:19 ~height:1 ~borderwidth:1 ~relief:`Groove frame_print_command2  in
	pack ~side:`Top ~fill:`Y [frame_print_command1];
	pack ~side:`Left ~fill:`Y [label_print_command];
	pack ~side:`Left ~fill:`Y [entry_print_command];
	pack ~side:`Top ~fill:`Y [frame_print_command2];
	pack ~side:`Left ~fill:`Y [button_ok_print_command];
	pack ~side:`Left ~fill:`Y [button_cancel_print_command];
	
) men;;
Menu.add_separator men;;
Menu.add_command ~label:"Quit" ~command:(fun () -> closeTk (); exit 1 ) men;;
Menubutton.configure ~menu:men meb;;
pack ~side:`Left [meb];;

(** Menu_button "Modes" *)
let meb_modes = (Menubutton.create ~text:"Modes" ~font:"guifont" ~activebackground:blue_light bar : Widget.menubutton Widget.widget);;
(** Menu "Modes" *)
let men_modes = (Menu.create meb_modes ~activebackground:blue_light ~borderwidth:1 ~relief:`Groove : Widget.menu Widget.widget);;
Menu.add_checkbutton ~label:"Untyped variables" 
	~onvalue:"1" ~offvalue:"0" ~variable:tv_untypedButton
	~command:(fun() ->
		click := (-1); 
		if(!untyped_variable) 
		then untyped_variable:=false 
		else untyped_variable:=true) men_modes;;
Menu.add_separator men_modes;;
(** Textvariable for the checkbutton "view messages emissions" *)
let tv_messageEmissionButton = (Textvariable.create () : Textvariable.textVariable);;
(* no emissions at initialisation *)
Textvariable.set tv_messageEmissionButton "0";;
Menu.add_checkbutton ~label:"view messages emissions" 
	~onvalue:"1" ~offvalue:"0" ~variable:tv_messageEmissionButton 
	~command:(fun() ->
		click := (-1); 
		if(!message_emission) 
			then (message_emission:=false ;  Canvas.configure_text canvas (`Tag "emission") ~state:(`Hidden) )
			else (message_emission:=true ;  Canvas.configure_text canvas (`Tag "emission") ~state:(`Disabled)) ) men_modes;;
Menubutton.configure ~menu:men_modes meb_modes;;
pack ~side:`Left [meb_modes];;
Menu.add_separator men_modes;;
(** Textvariable for the checkbutton "view sender pattern message" *)
let tv_messageSender = (Textvariable.create () : Textvariable.textVariable);;
Textvariable.set tv_messageSender "0";;
Menu.add_checkbutton ~label:"view sender pattern message" 
	~onvalue:"1" ~offvalue:"0" ~variable:tv_messageSender
	~command:(fun() ->
		click := (-1); 
		if(!message_sender) 
			then (message_sender:=false ;  Canvas.configure_text canvas (`Tag "message_sender") ~state:(`Hidden) )
			else (message_sender:=true ;  Canvas.configure_text canvas (`Tag "message_sender") ~state:(`Disabled)) ) men_modes;;
Menubutton.configure ~menu:men_modes meb_modes;;
pack ~side:`Left [meb_modes];;
Menu.add_separator men_modes;;
(** Textvariable for the checkbutton "view receiver pattern message" *)
let tv_messageReceiver = (Textvariable.create () : Textvariable.textVariable);;
Textvariable.set tv_messageReceiver "0";;
Menu.add_checkbutton ~label:"view receiver pattern message" 
	~onvalue:"1" ~offvalue:"0" ~variable:tv_messageReceiver
	~command:(fun() ->
		click := (-1); 
		if(!message_receiver) 
			then (message_receiver:=false ;  Canvas.configure_text canvas (`Tag "message_receiver") ~state:(`Hidden) )
			else (message_receiver:=true ;  Canvas.configure_text canvas (`Tag "message_receiver") ~state:(`Disabled)) ) men_modes;;
Menubutton.configure ~menu:men_modes meb_modes;;
pack ~side:`Left [meb_modes];;
Menu.add_separator men_modes;;
(** Textvariable for the checkbutton "view real message" *)
let tv_messageReal = (Textvariable.create () : Textvariable.textVariable);;
Textvariable.set tv_messageReal "1";;
(** Textvariable for the checkbutton "view sender pattern in principal position" *)
let tv_principalSender = (Textvariable.create () : Textvariable.textVariable);;
Textvariable.set tv_principalSender "0";;
Menu.add_checkbutton ~label:"view real message" 
	~onvalue:"1" ~offvalue:"0" ~variable:tv_messageReal
	~command:(fun() ->
		click := (-1); 
		if(!message_real) 
			then (message_real:=false ;  Canvas.configure_text canvas (`Tag "message_real") ~state:(`Hidden) )
			else (message_real:=true ; message_principal_sender:=false ; Textvariable.set tv_principalSender "0";  Canvas.configure_text canvas (`Tag "message_principal_sender") ~state:(`Hidden) ; Canvas.configure_text canvas (`Tag "message_real") ~state:(`Disabled)) ) men_modes;;
Menubutton.configure ~menu:men_modes meb_modes;;
pack ~side:`Left [meb_modes];;
Menu.add_separator men_modes;;
Menu.add_checkbutton ~label:"view sender pattern in principal position" 
	~onvalue:"1" ~offvalue:"0" ~variable:tv_principalSender
	~command:(fun() ->
		click := (-1);
		if (!message_principal_sender) 
			then (message_principal_sender:=false ; Canvas.configure_text canvas (`Tag "message_principal_sender") ~state:(`Hidden) )
			else (message_principal_sender:=true ; message_real:=false ; Textvariable.set tv_messageReal "0"; Canvas.configure_text canvas (`Tag "message_real") ~state:(`Hidden) ; Canvas.configure_text canvas (`Tag "message_principal_sender") ~state:(`Disabled)) ) men_modes;;
Menubutton.configure ~menu:men_modes meb_modes;;
pack ~side:`Left [meb_modes];;
Menu.add_separator men_modes;;
(** for change fonts size *)
Menu.add_command ~label:"Font size..." ~command:(fun () ->
	let string_to_int s = (try 
    				int_of_string s
  				with _ -> -1 ) in
	let file_config_command = (Utils_simulator.path_span^"/bin/protocol_simulation/font_size.ini") in

	let root_font_size =  Toplevel.create ~borderwidth:1 ~relief:`Groove root in
	Wm.title_set root_font_size ("font size ...");
	let frame_font_size1 = Frame.create ~borderwidth:1 ~relief:`Sunken root_font_size in
	let frame_font_size2 = Frame.create ~borderwidth:1 ~relief:`Sunken root_font_size in
	let label_font_size = Label.create ~text:"Font size : " ~width:15 ~height:2 frame_font_size1 in
	let entry_font_size = Entry.create ~width:20 ~borderwidth:1 frame_font_size1 ~background:selcol in
	Entry.insert entry_font_size ~index:`End ~text:(string_of_int (!current_font_size)) ;
	let button_ok_font_size = 	Button.create ~text:"OK" 
						~command:(fun () -> 	let tmp_size =  
    										string_to_int (Entry.get entry_font_size)
									in
									if(tmp_size>0 && tmp_size<>(!current_font_size))
									then (	if(Sys.file_exists file_config_command) 
										then ignore(Sys.remove file_config_command);
										current_font_size := tmp_size;
										set_size_of_all_fonts tmp_size;
										saveFile file_config_command (string_of_int tmp_size)
										);
										
									destroy label_font_size ; destroy entry_font_size ; 
									destroy entry_font_size ; destroy root_font_size) 
					~width:19 ~height:1 ~borderwidth:1 ~relief:`Groove frame_font_size2  in
	let button_cancel_font_size = Button.create ~text:"Cancel" 
						~command:(fun () -> 	destroy label_font_size ; destroy entry_font_size ; 
									destroy button_ok_font_size ; destroy entry_font_size ; 
									destroy root_font_size ) 
						~width:19 ~height:1 ~borderwidth:1 ~relief:`Groove frame_font_size2  in
	pack ~side:`Top ~fill:`Y [frame_font_size1];
	pack ~side:`Left ~fill:`Y [label_font_size];
	pack ~side:`Left ~fill:`Y [entry_font_size];
	pack ~side:`Top ~fill:`Y [frame_font_size2];
	pack ~side:`Left ~fill:`Y [button_ok_font_size];
	pack ~side:`Left ~fill:`Y [button_cancel_font_size];
	
) men_modes;;
Menubutton.configure ~menu:men_modes meb_modes;;
pack ~side:`Left [meb_modes];;


(** Menu_button "Variables monitoring" *)
let meb_var = (Menubutton.create ~text:"Variables monitoring" ~font:"guifont" ~activebackground:blue_light bar : Widget.menubutton Widget.widget);;
(** Menu "Variables monitoring" *)
let men_var = (Menu.create meb_var ~activebackground:blue_light ~borderwidth:1 ~relief:`Groove : Widget.menu Widget.widget);;
Menubutton.configure ~menu:men_var meb_var;;
pack ~side:`Left [meb_var];;

(** MSC label *)
let meb_msc_label = (Menubutton.create ~text:" MSC " ~font:"guifont" bar : Widget.menubutton Widget.widget);;
pack ~side:`Right [meb_msc_label];;

(* ******************************************************* *) 



(*

   \subsection{Main function}
  
   The [print_if] function is the main Intermediate Format printing function. 
   It displays IF sections according the the command line arguments. 
   By default, this function displays all sections.
  \begin{center}
  type:
   [val print_if : role_instance list -> Types.goals_type list -> unit]
  \end{center}
*)
(** Principal of this modul call in main.ml *)
let print_if (role_instances : role_instance list) (goals : Types.goals_type list) : unit =

  (* Title setting *)
  Wm.title_set root (sim_version^" : "^(Filename.basename (!Globals.input_file)));
  fileName := (!Globals.input_file);

  if not (!Globals.flag_stdout || !Globals.flag_split) then
    Globals.tmp_channel := open_out !Globals.tmp_file;
  if not (!Globals.flag_split) then
    prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");

 

      print_section_signatureA role_instances;
      print_section_types ();
      print_section_init role_instances;
      automate:=print_section_rulesA role_instances [] [] [];

      print_type_in_role_session (!automate);
      
      let symplify_automate () =
       let fusion_final_state (s1 : string * string) (s2 : string * string) =
 		let decomp1 = (cdr(decompose ("x("^(findState s1)^")"))) in
 		let decomp2 = (cdr(decompose ("x("^(findState s2)^")"))) in
		let rec aux l1 l2 res = 
			match l1 with [] ->	let tmp = List.rev res in 
						if (tmp=[]) then ""
						else (List.fold_left (fun a b -> a^","^b) (car tmp) (cdr tmp))
				|(x1::xs1) -> 	match l2 with [] ->  	let tmp = List.rev res in 
									if (tmp=[]) then ""
									else (List.fold_left (fun a b -> a^","^b) (car tmp) (cdr tmp))
							|(x2::xs2) -> 	if(is_a_refresh_var x1 && not(is_a_refresh_var x2)) 
									then aux xs1 xs2 ((x2^"_new")::res)
									else aux xs1 xs2 (x2::res)
		
		in ((aux decomp1 decomp2 []),(findIknows s2))
	in
      let rec remove_test m =
 		if(m="") then ""
		else(
			if(is_a_test_var m) then (get_name_of_test_var m)
			else if((is_a_value m) || (is_a_var m)) then m
			else let l1= decompose m in
			     if(l1<>[])
			     then (
				let rec aux ll1  =
					 match ll1 with [] -> ""
						|[x] ->(remove_test x)
						|_ ->((remove_test (car ll1))^","^(aux (cdr ll1)))
				in
				(car l1)^"("^(aux (cdr l1))^")"; 
				)
			     else ""	
		)
	in     
      let rec assign_test_var_name m test_var_list=
		if(m="") then ""
		else(
			if((List.mem m test_var_list)) then ("Test_"^m)
			else if((is_a_value m) || (is_a_var m)) then m
			else let l1= decompose m in
			     if(l1<>[])
			     then (
				let rec aux ll1  =
					 match ll1 with [] -> ""
						|[x] ->(assign_test_var_name x test_var_list)
						|_ ->((assign_test_var_name (car ll1) test_var_list)^","^(aux (cdr ll1)))
				in
				(car l1)^"("^(aux (cdr l1))^")"; 
				)
			     else ""	
		)
	in
       let treatment state iknows =
		let state_d = (cdr (decompose ("x("^state^")"))) in
		let test_var_list_ref = ref [] in
		let new_state = List.map (fun v ->	if (is_a_test_var v)
							then (
							if(List.mem (get_name_of_test_var v) (!test_var_list_ref)) then "X_X" else (get_name_of_test_var v))
							else 	if ((is_a_var v) (*&& not( (Hashtbl.find role_signature v) = ["hash_func"] || (Hashtbl.find role_signature v) = ["function"]  || (decompose v)<>[] )*)) 
								then (test_var_list_ref:=v::(!test_var_list_ref);v)
							else v)
				state_d in
		let new_state2 = (List.fold_left (fun a b -> a^","^b) "" new_state) in
		let new_state3 = String.sub new_state2 1 ((String.length new_state2)-1) in
		let new_iknows = assign_test_var_name iknows (!test_var_list_ref) in
		(new_state3,new_iknows)
	in
 	let simplify_var name id actionsL = 
		List.map 
		(fun (initial, operation, (finalState, finalIknows)) -> 
		(	(treatment (findState initial)(findIknows initial)), 
			((List.map remove_test (fst (fst operation)), List.map remove_test (snd (fst operation))), (List.map remove_test (fst (snd operation)), List.map remove_test (snd (snd operation)))) , 
			(finalState,(remove_test finalIknows)))
		)
		actionsL
	in
      
	let zip_operation l_op1 lop2 =
		let zip l1 l2 = remove_double l1@l2 in
		(
			( 
				(zip (fst (fst l_op1))(fst (fst lop2))) , (zip (snd (fst l_op1))(snd (fst lop2))) 
			), 
			(
				(zip (fst (snd l_op1))(fst (snd lop2))) , (zip (snd (snd l_op1))(snd (snd lop2)))
			)
		) 
	in
        let res = ref [] in
	let buff = ref [] in 
	let add_in_res x = 	if((List.mem x (!res))
					|| (List.mem x (!buff)))
				then () 
				else (res:= x::(!res)) in 
	let add_in_buff x = if(List.mem x (!buff)) then () else (buff:= x::(!buff)) in 
	let symplify_actions_aux name id actionsL =
		let find_state state =
			let rec aux l res =
				match l with [] -> List.rev res
					|((initial,operation,final)::ls) -> 	
									if( (replace "Test_" "" (findState initial))=(replace "Test_" "" state) )
									then aux ls ((initial,operation,final)::res)
									else aux ls res;
			in
			aux actionsL []
		in			
		let name_in_session = get_name_in_session name id in
		res:=[];
		buff:=[];
		List.iter 
		(fun (initial,operation,final) -> 
			if((findIknows final)<> "") then  add_in_res (initial,operation,final)
			else (List.iter (fun (initial2,operation2,final2) -> 
					if((findIknows initial2)<> "")
					then (	add_in_res (initial,operation,final); add_in_res (initial2,operation2,final2);
						List.iter 
						(fun s -> 
							if(not(List.mem s (!res))) then res:=s::(!res))
						(find_state (findState initial2)))
					else (if (match_state name_in_session (findState final) name_in_session (findState initial2)) 
						then (	 
							add_in_buff (initial,operation,final);
							add_in_buff (initial2,operation2,final2);
							add_in_res (initial,(zip_operation operation operation2),(fusion_final_state final final2)))
						else (	add_in_res (initial,operation,final);
							add_in_res (initial2,operation2,final2)););		
					)actionsL))
			actionsL;
			!res
	in
	let symplify_actions_aux2 name id actionsL =
		let name_in_session = get_name_in_session name id in
		res:=[];
		buff:=[];
		List.iter 
		(fun (initial,operation,final) -> 
			if((findIknows final)<> "") then  add_in_res (initial,operation,final)
			else (List.iter (fun (initial2,operation2,final2) -> 
				(List.iter (fun (initial3,operation3,final3) -> 
			
					if((findIknows initial2)<> "" || (findIknows initial2)<> "")
					then (add_in_res (initial,operation,final); add_in_res (initial2,operation2,final2))
					else (if (match_state name_in_session (findState final) name_in_session (findState initial3)) 
						then (	 
							add_in_buff (initial,operation,final);
							add_in_buff (initial2,operation2,final2);
							add_in_buff (initial3,operation3,final3);
							add_in_res (initial,(zip_operation (zip_operation operation operation2) operation3),(fusion_final_state (fusion_final_state final final2) final3)))
						else (	add_in_res (initial,operation,final);
							add_in_res (initial2,operation2,final2);
							add_in_res (initial3,operation3,final3)))	

						)actionsL)								
					)actionsL))
			actionsL;
			!res
	in
	let rec simplify_actions2 name id actionsL length =
		let tmp = remove_double (symplify_actions_aux2 name id actionsL)  in
		let length2 = List.length tmp in
		if (length<>length2) then (simplify_actions2 name id tmp length2)
		else tmp;
	in
	let rec simplify_actions name id actionsL length =
		let tmp = remove_double (symplify_actions_aux name id actionsL) in
		let length2 = List.length tmp in
		if (length<>length2) then (simplify_actions name id tmp length2)
		else (simplify_actions2 name id tmp length);
	in
	let simplify_operations actionsL =
		let place_delete_test_aux l1_test l2 =
		remove ""
		(List.map
			(fun e ->
			let d = decompose e in 
			if(d<>[] && (car d)="contains")
			then( 
				if(	not(List.mem e (fst l2)) 
					&& (List.mem ("delete"^"("^(list2string "," (cdr d))^")") (snd l1_test) ) )
				then ""
				else e
			)
			else e
			)
		(fst l1_test) )
		in
		let place_delete_test l1 l2 = 
			(( (place_delete_test_aux (fst l1) (fst l2)), snd(fst l1)),snd l1)
		in
		let rec aux initial1 operations1 final1 actionsL2 =
			match actionsL2 with [] -> (initial1,operations1,final1)
					|((initial2,operations2,final2)::ls) -> if( operations1<>operations2 && initial1=initial2 && final1=final2 )
										then aux initial1 (zip_operation (place_delete_test operations1 operations2) (place_delete_test operations2 operations1) ) final1 ls
										else  aux initial1 operations1 final1 ls
	
		in 
		let rec simplify actionsL res buff =
			match actionsL with [] -> List.rev res
					|((initial1,operations1,final1)::ls) -> if(List.mem (initial1,final1) buff)
										then simplify ls res buff
										else simplify ls ((aux initial1 operations1 final1 actionsL)::res) ((initial1,final1)::buff) 
			in simplify actionsL [] []
	in
	multi_t := 
		List.map (fun (name, id, actionsL) ->
				let actions_with_simplify_operations = simplify_operations actionsL in
				(name, id, (simplify_var name id (simplify_actions name id actions_with_simplify_operations (List.length actions_with_simplify_operations))))
		) (!multi_t);
	automate:=
	(List.map
	(fun (name,id,statesL,actionsL) -> 
		let actions_with_simplify_operations = simplify_operations actionsL in
	(*
		(List.iter 
	(fun (initial,((tg,ag),(td,ad)),final) -> prerr_string ((findState initial)^" - "^(findIknows initial)^" [ "^(list2string "," tg)^" - "^(list2string "," ag)^" - "^(list2string "," td)^" - "^(list2string "," ad)^" ] =|>\n"^(findState final)^" - "^(findIknows final)^"\n\n"))  
	(simplify_var name id (simplify_actions name id actions_with_simplify_operations (List.length actions_with_simplify_operations))));
	*)
	
	(*
	 		(List.iter 
	(fun (initial,((tg,ag),(td,ad)),final) -> prerr_string ((findState initial)^" - "^(findIknows initial)^" [ "^(list2string "," tg)^" - "^(list2string "," ag)^" - "^(list2string "," td)^" - "^(list2string "," ad)^" ] =|>\n"^(findState final)^" - "^(findIknows final)^"\n\n")) 
	(actions_with_simplify_operations)); 
	*)
	(*
		 		(List.iter 
	(fun (initial,((tg,ag),(td,ad)),final) -> prerr_string ((findState initial)^" - "^(findIknows initial)^" [ "^(list2string "," tg)^" - "^(list2string "," ag)^" - "^(list2string "," td)^" - "^(list2string "," ad)^" ] =|>\n"^(findState final)^" - "^(findIknows final)^"\n\n")) 
	(actionsL)); 
		*)
		(name,id,statesL,(simplify_var name id (simplify_actions name id actions_with_simplify_operations (List.length actions_with_simplify_operations))))
		)
	(!automate)) in
	symplify_automate();

      
(* **** *)
      
      if(!intruder_mode)
      then(

      	pack ~side:`Left ~fill:`Y [intruder_knowledge_v];
	pack ~side:`Bottom ~fill:`X [intruder_knowledge_h];	
	pack ~side:`Left ~fill:`Both ~expand:true [label_intruder];
	pack ~side:`Left ~fill:`Both ~expand:true [window_compose_intruder_knowledge];
	pack ~side:`Top ~fill:`Both ~expand:false [intruderFrameForButton];	
	pack ~side:`Top ~fill:`Both ~expand:true [intruder_knowledge_lst];
	pack ~side:`Top ~fill:`Both ~expand:true [intruderFrame];
      
	add_type "(Intruder_, 0)" ["agent" ; "set(message)" ; "message"];
	
	add_type "intruder_" ["agent"];
	add_type "set_knowledge" ["set(message)"];
	add_type "dummy_msg" ["message"];
	add_type "0" ["nat"];
	add_type "Intruder_" ["agent"];
	add_type "Set_knowledge" ["set(message)"];
	add_type "Listen_i" ["message"];
	
      	add_state  "(Intruder_, 0)" "intruder_,0,set_knowledge,dummy_msg";
	
	let intruder_role = ("Intruder_", 0, ref [], [("Intruder_,0,Set_knowledge,Listen_i","Listen_i"), (([],[]), ([],["cons(Listen_i,Set_knowledge)"])), ("Intruder_,0,Set_knowledge,Listen_i","")]) in
      	automate := intruder_role::(!automate);
      );

(* *** *)


  List.iter
   (fun (name,id,statesL,actionsL) ->
      if ((!statesL)=[])
      
      then(
	   let name_in_session = get_name_in_session name id in
	   let initial_state = get_initial_state name_in_session in 
	   let not_contain_the_start_command = not(List.mem "start" (List.map (fun (initial,_,_) -> findIknows initial) actionsL)) in
	   List.iter
           (fun (initial,_,final) ->
	     if(match_state name_in_session initial_state name_in_session (findState initial))
	     then(
		List.iter
		(fun t ->
			if ((is_a_test_var t) 
			&& not(List.mem (get_name_of_var t) (Hashtbl.find_all role_var_monitoring name_in_session))) 
			then Hashtbl.add role_var_monitoring name_in_session (get_name_of_var t)
			else if ((t<>"X_X") && (is_a_var t) 
				&& not(List.mem t (Hashtbl.find_all role_var_monitoring name_in_session))) 
			then Hashtbl.add role_var_monitoring name_in_session t)
		(cdr(decompose ("x("^(findState initial)^")")));
		
		if((findIknows initial)="start")
		  then (ignore(assign_var_of_state name_in_session initial_state name_in_session (findState initial));
		  	statesL:=((findState final), ref true)::(!statesL) ;  
		  	(*refresh_var_of_state name_in_session (findState final) (!tag_number);*)  
			operations_affectation_treatment name_in_session (findState final);
			)
		  else (if not_contain_the_start_command then ignore(assign_var_of_state name_in_session initial_state name_in_session (findState initial));
		  	statesL:=((findState initial), ref false)::(!statesL) ;  
			add_in_rcv_stack name_in_session (findState initial) (findIknows initial);
		  	(*refresh_var_of_state name_in_session (findState final) (!tag_number);  *)
			operations_affectation_treatment name_in_session (findState final);
			)
		  ))
           actionsL
          )
   )
  (!automate);
  
  if(!intruder_mode)
	then(
		List.iter
			(fun k -> add_in_set "(Intruder_, 0)" k "(Intruder_, 0)" "Set_knowledge")
		!intruder_knowledge;
  		update_intruder_knowledge ();
		display_intruder_knowledge();
	);	

	     	  (*   (List.iter
	(fun (name,id,statesL,actionsL) -> 
		(List.iter 
	(fun (initial,((tg,ag),(td,ad)),final) -> prerr_string ((findState initial)^" - "^(findIknows initial)^" [ "^(list2string "," tg)^" - "^(list2string "," ag)^" - "^(list2string "," td)^" - "^(list2string "," ad)^" ] =|>\n"^(findState final)^" - "^(findIknows final)^"\n\n"))  
	actionsL);
	) (!automate)); *)
	
  xinit_coord (List.length (!automate));
  let x_pos = ref 0. in
    List.iter
      (fun (name,id,_,_) ->
	let name_in_session = (get_name_in_session name id) in
      		Menu.add_command ~label:name_in_session ~command:(fun () -> var_frame name_in_session) men_var;
		draw_plot (!x_pos) name_in_session (if(name_in_session="(Intruder_, 0)" && !intruder_mode) then red else blue);
                draw_plot (!x_pos) name_in_session (if(name_in_session="(Intruder_, 0)" && !intruder_mode) then red else blue);
	       x_pos:= 10. +. (!x_pos))
      (!automate);

     (* view_table_var ();*)
      
  add_in_automate_story ();
  add_in_variable_story ();
  next_message_stack ();
  next_id_refresh_story ();
  insert_lst "" "";
  
  
(* **** *)


  if not (!Globals.flag_stdout || !Globals.flag_split) then
    Globals.close_out_noerr !Globals.tmp_channel;

  (* 
     Indicating the total number of warnings, if any.
  *)
  if(Sys.file_exists (!Globals.tmp_file)) then ignore(Sys.remove (!Globals.tmp_file));
  mainLoop();
  if (!Error_handler.warningsCounter <> 0) then
    (flush stdout;
     prerr_string "\n%% Number of warnings: "; 
     prerr_int !Error_handler.warningsCounter; 
     prerr_newline();
     flush stderr)
;;


